// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - May 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// Define this to help debug the client by making client/server log
// file and so on. TODO: Make this a registry setting?
// #define CLIENT_DEBUG

#ifdef _MSC_VER
#pragma warning(disable:4786 4503)
#endif

#include "CVSAction.h"
#include "CVSStatus.h"
#include "../Utils/PathUtils.h"
#include "../Utils/TortoiseUtils.h"
#include "../Utils/TortoiseRegistry.h"
#include <fstream>
#include "../TortoiseAct/TortoiseDialogs.h"
#include "../TortoiseAct/TortoiseTip.h"
#include "../Utils/Keyboard.h"

#include "../Utils/TortoiseDebug.h"

#include "../TortoiseAct/ThreadSafeProgress.h"

#include <io.h>
#include <fcntl.h>

bool CVSAction::ourInUse = false;
std::string CVSAction::ourProtocol;
std::string CVSAction::ourUserName;
std::string CVSAction::ourServerMachine;
std::string CVSAction::ourPassword;
std::string CVSAction::ourRepositoryDirectory;
std::string CVSAction::ourCVSROOT;
std::string CVSAction::ourStreamingToFile;
std::string CVSAction::ourErrorStore;
std::string CVSAction::ourOutputStore;

ThreadSafeProgress* CVSAction::ourThreadSafeProgressDialog;
DWORD CVSAction::ourLastTick;

CVSActionPipe * CVSAction::consoleout_pipe = 0;

bool CVSAction::PipeToGUI(CVSActionPipe *pPipe, std::string &store)
{
    std::string text = pPipe->TextOut();
    int length = text.length();

    if (length > 0)
    {
        ourThreadSafeProgressDialog->NewText(
            const_cast<char*>(text.c_str()), length);
        store += text;
        if (consoleout_pipe)
            fwrite(text.c_str(), 1, length, consoleout_pipe->InputFile());
    }

    return (length > 0);
}

DWORD WINAPI CVSAction::PipeThread(LPVOID pActionPipe)
{
    CVSActionPipe *pPipe = reinterpret_cast<CVSActionPipe*>(pActionPipe);

    const int bufSize = 128 * 1024; // 128Kb
    char readBuffer[bufSize];
    DWORD bytes = 0;
    while (
        ReadFile(pPipe->OutputHandle(), readBuffer, bufSize - 1, &bytes, NULL) == TRUE)
    {
        readBuffer[bytes] = '\0';
        std::string message = readBuffer;
        if (message.length() > 0)
            pPipe->TextIn(message);
    }

    return 0;
}


void CVSAction::SetEnvironment(std::string name, std::string value)
{
    std::string entry = name + "=" + value;
    putenv(entry.c_str());
}

CVSAction::CVSAction(int flags)
{
	if (flags & PIPE_STDOUT)
		consoleout_pipe = new CVSActionPipe(CVSActionPipe::INPUT_IS_FILE + CVSActionPipe::OUTPUT_IS_FILE);
	else
		consoleout_pipe = 0;

	myProgressDialog = CreateProgressDialog((flags & HIDE_PROGRESS) == 0);
	ourThreadSafeProgressDialog = new ThreadSafeProgress(myProgressDialog);

	ourLastTick = GetTickCount();
	ourStreamingToFile = "";

	myCloseIfOK = TortoiseRegistry::ReadBoolean("Close If No Error", true);
	if (IsControlDown())
		myCloseIfOK = false;
	myCloseAnyway = false;

	myUses = 0;
	myErrors = 0;

	myThread.BeginThread(this, &CVSAction::ThreadMain);
}

void CVSAction::SetStreamFile(const std::string& file)
{
	ourStreamingToFile = file;
	DeleteFile(ourStreamingToFile.c_str());
}

void CVSAction::SetCloseIfOK(bool closeIfOK)
{
	myCloseIfOK = closeIfOK;
}

void CVSAction::SetCloseAnyway(bool closeAnyway)
{
	myCloseAnyway = closeAnyway;
}

bool CVSAction::Command(std::string currentDirectory, MakeArgs& args)
{
	bool ret;
	bool repeat;
	do
	{
		repeat = false;
		ret = DoCommand(currentDirectory, args);

		// Look for instructions to run "cvs login",
		// and ask for their password
		if (ourErrorStore.find("\"cvs login\"", 0) != std::string::npos
			|| ourErrorStore.find("authorization failed:", 0) != std::string::npos)
		{
			--myErrors; // not a real error, as we hook it!
			std::string password;
			if (DoLoginDialog(password))
			{
				MakeArgs loginArgs;
				loginArgs.add("login");
				ourPassword = password;

				ret = DoCommand(currentDirectory, loginArgs);
				if (!ret)
					--myErrors;
				repeat = true;
			}
			else
			{
				std::string message =

				"Tortoise Tip:  Sometimes login can fail if you have the repository directory\n"
				"wrong in the Checkout dialog.  Make sure you have a slash at the start of the\n"
				"directory, and that the case and path are correct.\n\n"
				"Examples are:  \"/cvsroot\", \"/usr/local/cvs-repository\", \"/cvsroot/cvsgui\"";
				
				myCloseIfOK = false;

				myProgressDialog->NewText(const_cast<char*>(message.c_str()), message.size());
			}

			// Delete streaming file (we don't want data from login etc.,
			// only from the operations performed successfully)
			if (!ourStreamingToFile.empty())
				DeleteFile(ourStreamingToFile.c_str());
		}
	}
	while (repeat);

	// Don't keep the password in memory any longer than necessary
	ourPassword = "";
	// Clear other stuff so values don't confuse if we reuse CVSAction later
	ourProtocol = "";
	ourUserName = "";
	ourServerMachine = "";
	ourRepositoryDirectory = "";
	ourCVSROOT = "";
	ourStreamingToFile = "";

	return ret;
}

bool CVSAction::DoCommandThread(std::string currentDirectory, MakeArgs args)
{
	// Prevent more than one at once, as the CVS DLL and hence
	// the progress dialog can't cope with it.
	if (ourInUse)
	{
		DoMessageDialog("Internal error, CVSAction::DoCommand used twice within same process.");
		return false;
	}
	ourInUse = true;

	if (ourCVSROOT.empty())
	{
		ourCVSROOT = CVSStatus::CVSRootForPath(currentDirectory);
		if (ourProtocol.empty())
			ourProtocol = CVSStatus::RootProtocol(ourCVSROOT);
		if (ourUserName.empty())
			ourUserName = CVSStatus::RootUserName(ourCVSROOT);
		if (ourServerMachine.empty())
			ourServerMachine = CVSStatus::RootServer(ourCVSROOT);
	}

	bool ret = false;

	// Store in static member, as unfortunately the callback
	// mechanism from the DLL doesn't have any way of passing
	// data to associate with a particular progress dialog
	// (I think)
	ourErrorStore = "";
	ourOutputStore = "";

	std::string whatDoing = "In " + currentDirectory + ": ";
	whatDoing += args.requoted() + "\n";
    ourThreadSafeProgressDialog->
        NewText(const_cast<char*>(whatDoing.c_str()),
                whatDoing.length());

	std::string oldError = ourErrorStore;
    int result = 0;

    HANDLE hWriteStdOut = NULL;
    CVSActionPipe *pErrPipe = new CVSActionPipe,
                  *pOutPipe = NULL;

    if (!ourStreamingToFile.empty())
    {
        hWriteStdOut = CreateFile(
            ourStreamingToFile.c_str(),
            GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE,
            NULL,
            CREATE_ALWAYS,
            FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH,
            NULL);
    }
    else
    {
        pOutPipe = new CVSActionPipe;
        hWriteStdOut = pOutPipe->InputHandle();
    }

    STARTUPINFO startupInfo;
    memset(&startupInfo, 0, sizeof(startupInfo));
    startupInfo.cb = sizeof(startupInfo);
    startupInfo.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
    startupInfo.hStdOutput = hWriteStdOut;
    startupInfo.hStdError = pErrPipe->InputHandle();
    startupInfo.wShowWindow = SW_HIDE;

    SetEnvironment("USER=", ourUserName);
    SetEnvironment("LOGNAME=", ourUserName);
    if (!ourCVSROOT.empty())
        SetEnvironment("CVSROOT", ourCVSROOT);
    SetEnvironment("CVS_GETPASS", ourPassword);
    if (ourProtocol != "server" && ourProtocol != "ext" && !ourServerMachine.empty())
        SetEnvironment("CVS_SERVER", ourServerMachine);
    if (ourProtocol == "ext")
        SetEnvironment("CVS_RSH", "ssh");
    if (TortoiseRegistry::ReadBoolean("Client Log", false))
        SetEnvironment("CVS_CLIENT_LOG",
                       GetTortoiseFolder() + "cvsclient.log");
    SetEnvironment("DIRTY_MCP", "no");
    std::string homeDir = TortoiseRegistry::ReadString("HOME");
    SetEnvironment("HOME", (homeDir.empty() ? "C:" : homeDir.c_str()));

    std::string arguments=args.requoted();
    PROCESS_INFORMATION processInfo;

    if (!CreateProcess(
        NULL,
        const_cast<char*>(args.requoted().c_str()),
        NULL,
        NULL,
        FALSE,
        0,
        NULL,
        const_cast<char*>(currentDirectory.c_str()),
        &startupInfo,
        &processInfo))
	{
		DoMessageDialog("Trouble launching the CVS process.\n\n"
						 "Please check there is a cvs.exe in the same\n"
						 "directory as TortoiseCVS or on the path.");
		// The rest of the code cleans up OK, even on this error,
		// so just fall through so all resources are freed properly.
	}

    DWORD threadID;
    HANDLE hOutThread = NULL, hErrThread = NULL;

    hErrThread = CreateThread(NULL, 0, PipeThread,
        reinterpret_cast<LPVOID>(pErrPipe), 0, &threadID);

    if (pOutPipe != NULL)
        hOutThread = CreateThread(NULL, 0, PipeThread,
            reinterpret_cast<LPVOID>(pOutPipe), 0, &threadID);

    bool textRead = false;
    DWORD exitCode = 0;
    do
    {
        Sleep(200);
        textRead = (PipeToGUI(pErrPipe, ourErrorStore) ||
                    (pOutPipe != NULL &&
                     PipeToGUI(pOutPipe, ourOutputStore)));
        GetExitCodeProcess(processInfo.hProcess, &exitCode);

    } while ((exitCode == STILL_ACTIVE || textRead) &&
             !ourThreadSafeProgressDialog->UserAborted());

    if (exitCode == STILL_ACTIVE)
    {
        TerminateProcess(processInfo.hProcess, 0);
        exitCode = 0;
    }
    result = exitCode;

    CloseHandle(processInfo.hProcess);

    if (WaitForSingleObject(hErrThread, 100) == WAIT_TIMEOUT)
        TerminateThread(hErrThread, 0);
    if (hOutThread != NULL &&
        WaitForSingleObject(hOutThread, 100) == WAIT_TIMEOUT)
        TerminateThread(hOutThread, 0);

    delete pErrPipe;
    if (pOutPipe != NULL)
        delete pOutPipe;
    else
        CloseHandle(hWriteStdOut);

	// Error codes for diff are stuffed (1 means there was a difference)
	// so we fudge our own more useful ones, based on if the CVS
	// operation wrote to stderr.
	if (args.main_command() == "diff")
	{
		if (oldError == ourErrorStore)
			result = 0;
		else
			result = 1;
	}

	myUses++;
	std::string message;
	if (result == 0)
	{
		ret = true;
		message = "Success, CVS operation completed";
	}
	else
	{
		message = "Error, CVS operation failed";
		myErrors++;
	}

	// Add the Tortoise tip
	std::string tip = TortoiseTip(ourOutputStore + ourErrorStore);
	if (!tip.empty())
		message += "\n\n" + tip;

	// Display the message
	message = "\n" + message + "\n\n";
    ourErrorStore += message;
    ourThreadSafeProgressDialog->NewText(
        const_cast<char*>(message.c_str()), message.length());

	ourInUse = false;

	return ret;
}

CVSAction::~CVSAction()
{
	// Check for control being pressed again (it is also checked
	// when the dialog is first made, so people can hold it down while
	// they select a menu item), and keep the dialog open if it is
	if (IsControlDown())
		myCloseIfOK = false;

	// If there were lots of operations, and one failed, remind the user
	if (myErrors > 0 && myUses > 1)
	{
		std::string message = "Error, one of the CVS operations failed";
        ourErrorStore += message;
        ourThreadSafeProgressDialog->NewText(
            const_cast<char*>(message.c_str()), message.length());
	}

	// bring to the front if something especially interesting to say
	if (myErrors > 0)
		myProgressDialog->Show(true);

	if ((myErrors > 0 || !myCloseIfOK) && !myCloseAnyway)
		myProgressDialog->WaitForAbort();

	if (myProgressDialog)
	{
		delete ourThreadSafeProgressDialog;
		ourThreadSafeProgressDialog = NULL;
		DeleteProgressDialog(myProgressDialog);
		myProgressDialog = NULL;
	}

	if (consoleout_pipe != 0)
		delete consoleout_pipe;
}

bool CVSAction::AtLeastOneSuccess()
{
	return (myErrors != myUses);
}

void CVSAction::SetCVSRoot(const std::string& protocol, const std::string& userName, const std::string& serverMachine, const std::string& repositoryDirectory)
{
	// Trailing slash on path name isn't liked
	std::string stippedDirectory = repositoryDirectory;
	RemoveTrailingDelimiter(stippedDirectory);

	ourProtocol = protocol;
	ourUserName = userName;
	ourServerMachine = serverMachine;
	ourRepositoryDirectory = stippedDirectory;
	
	ourCVSROOT = CalcCVSRoot(protocol, userName, serverMachine, stippedDirectory);
}
	
void CVSAction::SetPassword(const std::string& password)
{
	ourPassword = password;
}

std::string CVSAction::CalcCVSRoot(const std::string& protocol, const std::string& userName, const std::string& serverMachine, const std::string& repositoryDirectory)
{
	std::string CVSROOT;
	if (!protocol.empty())
		CVSROOT += ":" + protocol + ":";
	if (!userName.empty())
		CVSROOT += userName + "@";
	if (!serverMachine.empty())
		CVSROOT += serverMachine + ":";
	CVSROOT += repositoryDirectory;

	return CVSROOT;
}


void CVSAction::ThreadMain()
{
	while (myThread.WaitForMoreWork())
	{
		myThreadResult = DoCommandThread(myThreadDirectory, myThreadArgs);
		ourThreadSafeProgressDialog->Finished();
	}
}

bool CVSAction::DoCommand(std::string currentDirectory, MakeArgs args)
{
	myThreadDirectory = currentDirectory;
	myThreadArgs = args;

	// New stuff for the thread!
	myThread.SignalNewWork();

	// Make sure the dialog responds, and wait until the thread is done
	ourThreadSafeProgressDialog->Main();

	return myThreadResult;
}

void CVSAction::CloseConsoleOutput()
{
	if (consoleout_pipe != 0)
	{
		delete consoleout_pipe;
		consoleout_pipe = 0;
	}
}

ProgressDialogAbstract* CVSAction::GetProgressDialog()
{
	return myProgressDialog;
}

//****************************************************************************

CVSActionPipe::CVSActionPipe(int _flags) :
	flags(_flags)
{
	CreatePipe(&outputHandle, &inputHandle, 0, 0);
	if (flags & INPUT_IS_FILE)
	{
		int fno = _open_osfhandle((long)inputHandle, 0);
		inputFile = _fdopen(fno, "w");
		setvbuf( inputFile, NULL, _IONBF, 0 );
	}
	else
		inputFile = 0;
	if (flags & OUTPUT_IS_FILE)
	{
		int fno = _open_osfhandle((long)outputHandle, 0);
		outputFile = _fdopen(fno, "r");
		setvbuf( outputFile, NULL, _IONBF, 0 );
	}
	else
		outputFile = 0;
    mutex = CreateMutex(NULL, FALSE, "TortoiseCVSMutex");
}

CVSActionPipe::~CVSActionPipe()
{
	CloseInput();
	if (outputFile != 0)
		fclose(outputFile);
	else
		CloseHandle(outputHandle);
    WaitForSingleObject(mutex, INFINITE);
    CloseHandle(mutex);
}

void CVSActionPipe::CloseInput()
{
	if (inputFile != 0)
	{
		fclose(inputFile);
		inputFile = 0;
		inputHandle = 0;
	}
	else if (inputHandle != 0)
	{
		CloseHandle(inputHandle);
		inputHandle = 0;
	}
}

void CVSActionPipe::TextIn(std::string textToAdd)
{
    WaitForSingleObject(mutex, INFINITE);
    text += textToAdd;
    ReleaseMutex(mutex);
}

std::string CVSActionPipe::TextOut()
{
    WaitForSingleObject(mutex, INFINITE);
    std::string retVal = text;
    text = "";
    ReleaseMutex(mutex);
    return retVal;
}

