// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - May 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifdef _MSC_VER
#pragma warning(disable:4786 4503)
#endif

#include "CVSStatus.h"
#include "CvsEntries.h"
#include "../Utils/PathUtils.h"
#include <sys/stat.h>
#include <time.h>
#include "TextBinary.h"
#include <fstream>

#include "../Utils/TortoiseDebug.h"

std::map<std::string, ENTNODE> CVSStatus::ourLastEntries;
std::map<std::string, EntnodeData*> CVSStatus::ourEntnodeCache;
time_t CVSStatus::ourEntnodeCacheTimeStamp;
std::string CVSStatus::ourEntnodeCacheDirectory;
bool CVSStatus::ourEntnodeErased = true;

time_t CVSStatus::ourRecursiveStatusCacheTimeStamp;
CVSStatus::FileStatus CVSStatus::ourRecursiveStatus;
std::string CVSStatus::ourRecursiveStatusFilename;

// This calls the GetEntriesUncached (below) which actually does the work.
// The caching speeds up directory listing display in Explorer in TortoiseCVS.
EntnodeData* CVSStatus::GetEntries(const std::string& filename)
{
	std::string directory = GetDirectoryAbove(filename);

	// Erase cache ever couple of seconds, or
	// if we change directory
	if (directory != ourEntnodeCacheDirectory
		|| time(NULL) > ourEntnodeCacheTimeStamp + 1)
	{
		time(&ourEntnodeCacheTimeStamp);
		ourEntnodeCache.clear();
		ourEntnodeCacheDirectory = directory;
		ourEntnodeErased = true;
	}

	// Look for entry in cache
	std::map<std::string, EntnodeData*>::iterator it = ourEntnodeCache.find(filename);
	if (it != ourEntnodeCache.end())
		return it->second;

	// Fill in cache with newly calculated value
	EntnodeData* data = GetEntriesUncached(filename);
	ourEntnodeCache[filename] = data;
	return data;
}

// This stuff is a bit fiddly - I've tried to write it without
// breaking into the CvsEntries file too much.  That class was
// designed for scanning a whole directory, rather than individual
// files like we do
EntnodeData* CVSStatus::GetEntriesUncached(const std::string& filename)
{
	std::string directory = GetDirectoryAbove(filename);

	if (!CVSFolderHere(directory))
	{
		if (!CVSFolderHere(filename))
			return NULL;
		// outer directory
	}

	std::string justfilename = ExtractLastPart(filename);
	if (justfilename == "Cvs" || justfilename == "CVS")
		return NULL;

	if (ourEntnodeErased)
	{
		ourLastEntries.clear();
		Entries_Open(ourLastEntries, directory.c_str());
		ourEntnodeErased = false;
	}

	struct stat sb;
	int statReturnCode = stat(filename.c_str(), &sb);
	if (statReturnCode != 0)
		return NULL;

	EntnodeData* data = Entries_SetVisited(directory.c_str(), ourLastEntries, justfilename.c_str(),
		sb, IsDirectory(filename.c_str()));

	return data;
}

CVSStatus::FileStatus CVSStatus::GetFileStatusRecursive(const std::string& filename)
{
	// Erase cache ever couple of seconds
	if (filename != ourRecursiveStatusFilename
		|| time(NULL) > ourRecursiveStatusCacheTimeStamp + 2)
	{
		ourRecursiveStatus = GetFileStatusRecursiveUncached(filename);
		ourRecursiveStatusFilename = filename;
		time(&ourRecursiveStatusCacheTimeStamp);
	}
	else
	{
		// Somewhere to put breakpoints in the debugger
		ASSERT(true);
	}

	return ourRecursiveStatus;
}

// Recursively scans directories to see if they have changed files in them
CVSStatus::FileStatus CVSStatus::GetFileStatusRecursiveUncached(const std::string& filename)
{
	FileStatus status = GetFileStatus(filename);

	if (IsDirectory(filename.c_str()) &&
		(status == STATUS_OUTERDIRECTORY || status == STATUS_INCVS))
	{
		std::vector<std::string> scan;
		bool changes = CVSStatus::RecursiveHasChangedFiles(filename);
		if (changes)
			return STATUS_CHANGED;
	}

	return status;
}

CVSStatus::FileStatus CVSStatus::GetFileStatus(const std::string& filename)
{
	// Don't check top level folders - otherwise Win2K would access _all_
	// drives (including floppy disks) from My Computer, which is annoying
	if (IsRootPath(filename.c_str()))
         return STATUS_NOWHERENEAR_CVS;

	// Check for special case of the outer directory
	if (IsDirectory(filename.c_str())
		&& CVSFolderHere(filename) && !CVSFolderHere(GetDirectoryAbove(filename)))
	{
		return STATUS_OUTERDIRECTORY;
	}

	EntnodeData* data = GetEntries(filename);
	if (!data)
		return STATUS_NOWHERENEAR_CVS;

	if (data->IsUnknown())
		return STATUS_NOTINCVS;

//	if(data->IsIgnored())

	if (!data->IsUnmodified())
	{
		if (data->NeedsMerge())
			return STATUS_CONFLICT;
		else
			return STATUS_CHANGED;
	}

	return STATUS_INCVS;
}

bool CVSStatus::IsFileBinary(const std::string& filename)
{
	if (IsDirectory(filename.c_str()))
		return false;

	std::string dir = GetDirectoryPart(filename);
	std::string file = ExtractLastPart(filename);

	bool result = false;
	kTextBinTYPE type = ::FileIsBinary(file.c_str(), dir.c_str());
	if (type == kFileIsOK)
		result = true;

	return result;
}

bool CVSStatus::IsFileText(const std::string& filename)
{
	if (IsDirectory(filename.c_str()))
		return false;

	std::string dir = GetDirectoryPart(filename);
	std::string file = ExtractLastPart(filename);

	bool result = false;
	kTextBinTYPE type = ::FileIsText(file.c_str(), dir.c_str());
	if (type == kFileIsOK)
		result = true;

	return result;
}


bool CVSStatus::InCVSHasSubdirs(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	if (!data)
		return false;
	return data->GetType() == ENT_SUBDIR;
}

std::string CVSStatus::CVSRootForPath(std::string path)
{
	// We must be a directory with a CVS folder
	path = GetDirectoryPart(path);
	EnsureTrailingDelimiter(path);
	if (!CVSFolderHere(path))
		return "";

	std::string root;
	std::string rootFile = path + "CVS/Root";
	std::ifstream in(rootFile.c_str(), std::ios::in);
	if (!in.good())
		return "";
	std::getline(in, root);

	return root;
}

std::string CVSStatus::RootServer(const std::string& cvsroot)
{
	std::string::size_type pos = cvsroot.find_first_of("@");
	if (pos == std::string::npos || pos >= cvsroot.size() - 1)
	{
		pos = cvsroot.find_first_of(":", 1);
		if (pos == std::string::npos || pos >= cvsroot.size() - 1)
			return "";
	}

	std::string endPart = cvsroot.substr(pos + 1);
	pos = endPart.find_first_of(":");
	if (pos == std::string::npos)
		return "";

	std::string server = endPart.substr(0, pos);

	return server;
}

std::string CVSStatus::RootProtocol(const std::string& cvsroot)
{
	std::string::size_type pos = cvsroot.find_first_of(":");
	if (pos == std::string::npos || pos >= cvsroot.size() - 1)
		return "";

	std::string endPart = cvsroot.substr(pos + 1);
	pos = endPart.find_first_of(":");
	if (pos == std::string::npos)
		return "";

	std::string protocol = endPart.substr(0, pos);

	return protocol;
}

std::string CVSStatus::RootUserName(const std::string& cvsroot)
{
	std::string::size_type pos = cvsroot.find_first_of(":", 2);
	if (pos == std::string::npos || pos >= cvsroot.size() - 1)
		return "";

	std::string endPart = cvsroot.substr(pos + 1);
	pos = endPart.find_first_of("@");
	if (pos == std::string::npos)
		return "";

	return endPart.substr(0, pos);
}

std::string CVSStatus::ServerPath(std::string path)
{
	// We must be a directory with a CVS folder
	path = GetDirectoryPart(path);
	EnsureTrailingDelimiter(path);
	if (!CVSFolderHere(path))
		return "";

	std::string serverPath;
	std::string repositoryFile = path + "CVS/Repository";
	std::ifstream in(repositoryFile.c_str(), std::ios::in);
	if (!in.good())
		return "";
	std::getline(in, serverPath);

	return serverPath;
}

void CVSStatus::RecursiveScan(std::string scanDirectory, std::vector<std::string>& modified, std::vector<std::string>& added, std::vector<std::string>& removed)
{
	if (!IsDirectory(scanDirectory.c_str()))
	{
		FileStatus status = GetFileStatus(scanDirectory);
		// TODO: We should detect "added" in this case
		// - currently it just appears as "modified"
		if (status == STATUS_CHANGED || status == STATUS_CONFLICT)
			modified.push_back(scanDirectory);
		return;
	}

	EnsureTrailingDelimiter(scanDirectory);
	std::map<std::string, ENTNODE> entries;

	bool result = Entries_Open(entries, scanDirectory.c_str());
	if (!result)
		return;
	
	for (std::map<std::string, ENTNODE>::const_iterator it = entries.begin();
		it != entries.end(); it++)
	{
		std::string filename = scanDirectory + it->first;

		if (IsDirectory(filename.c_str()))
		{
			RecursiveScan(filename, modified, added, removed);
		}
		else
		{
			std::string justfilename = ExtractLastPart(filename);
			std::string directory = GetDirectoryAbove(filename);
			struct stat sb;
			memset(&sb, 0, sizeof(sb));
			int statReturnCode = stat(filename.c_str(), &sb);
			bool notThere = (statReturnCode != 0);

			EntnodeData* data = Entries_SetVisited(directory.c_str(), entries, justfilename.c_str(),
				sb, IsDirectory(filename.c_str()));

			if (!data->IsUnmodified())
			{
				if ((*data)[EntnodeFile::kVN] == std::string("0"))
					added.push_back(filename);
				else if (!notThere)
					modified.push_back(filename);
			}

			if (data->IsRemoved())
				removed.push_back(filename);
		}
	}

}

bool CVSStatus::RecursiveHasChangedFiles(std::string scanDirectory)
{
	if (!IsDirectory(scanDirectory.c_str()))
		return false;

	EnsureTrailingDelimiter(scanDirectory);
	std::map<std::string, ENTNODE> entries;

	bool result = Entries_Open(entries, scanDirectory.c_str());
	if (!result)
		return false;
	
	for (std::map<std::string, ENTNODE>::const_iterator it = entries.begin();
		it != entries.end(); it++)
	{
		std::string filename = scanDirectory + it->first;

		if (IsDirectory(filename.c_str()))
		{
			if (RecursiveHasChangedFiles(filename))
				return true;
		}
		else
		{
			std::string justfilename = ExtractLastPart(filename);
			std::string directory = GetDirectoryAbove(filename);
			struct stat sb;
			memset(&sb, 0, sizeof(sb));
			int statReturnCode = stat(filename.c_str(), &sb);
			bool notThere = (statReturnCode != 0);

			EntnodeData* data = Entries_SetVisited(directory.c_str(), entries, justfilename.c_str(),
				sb, IsDirectory(filename.c_str()));

			if (!data->IsUnmodified())
			{
				if ((*data)[EntnodeFile::kVN] == std::string("0"))
					return true;
				else if (!notThere)
					return true;
			}

			if (data->IsRemoved())
				return true;
		}
	}
	return false;
}

bool CVSStatus::HasRevisionNumber(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return(data != 0 && (*data)[EntnodeFile::kVN] != 0);
}

bool CVSStatus::HasTimestamp(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return(data != 0 && (*data)[EntnodeFile::kTS] != 0);
}

bool CVSStatus::HasStickyTag(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return(data != 0 && (*data)[EntnodeFile::kTagOnly] != 0);
}

bool CVSStatus::HasStickyDate(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return(data != 0 && (*data)[EntnodeFile::kDateOnly] != 0);
}

bool CVSStatus::HasOptions(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return(data != 0 && (*data)[EntnodeFile::kOption] != 0);
}

std::string CVSStatus::GetRevisionNumber(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return((*data)[EntnodeFile::kVN]);
}

std::string CVSStatus::GetTimestamp(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return((*data)[EntnodeFile::kTS]);
}

std::string CVSStatus::GetStickyTag(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return((*data)[EntnodeFile::kTagOnly]);
}

std::string CVSStatus::GetStickyDate(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return((*data)[EntnodeFile::kDateOnly]);
}

std::string CVSStatus::GetOptions(const std::string& filename)
{
	EntnodeData* data = GetEntries(filename);
	return((*data)[EntnodeFile::kOption]);
}
