/*
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 1, or (at your option)
** any later version.

** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.

** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * Author : Alexandre Parenteau <aubonbeurre@hotmail.com> --- April 1998
 */

/*
 * CvsEntries.cpp --- adaptation from cvs/src/entries.c
 */

#ifdef _MSC_VER
#pragma warning(disable:4786 4503)
#endif

#include "../Utils/TortoiseDebug.h"

#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>

#include "CvsEntries.h"
#include "getline.h"
#include "FileTraversal.h"
#include <stdio.h>
#include <string>
#include "../Utils/PathUtils.h"

#if TIME_WITH_SYS_TIME
#	include <sys/time.h>
#	include <time.h>
#elif HAVE_SYS_TIME_H
#	include <sys/time.h>
#else
#	include <time.h>
#endif

EntnodeData::~EntnodeData()
{
	--ref;
	if(user != 0L)
		free(user);
}

void EntnodeData::SetDesc(const char *newdesc)
{
	desc = newdesc;
}

EntnodeFile::~EntnodeFile()
{
	if(vn != 0L)
		free(vn);
	if(ts != 0L)
		free(ts);
	if(option != 0L)
		free(option);
	if(tag != 0L)
		free(tag);
	if(date != 0L)
		free(date);
	if(ts_conflict != 0L)
		free(ts_conflict);
}

void NoMem()
{
	exit(-1);
}

EntnodeFile::EntnodeFile(const char *newfullpath, const char *newuser, const char *newvn,
	const char *newts, const char *newoptions, const char *newtag,
	const char *newdate, const char *newts_conflict) : EntnodeData(ENT_FILE)
{
	vn = 0L;
	ts = 0L;
	option = 0L;
	tag = 0L;
	date = 0L;
	ts_conflict = 0L;

	if(newuser != 0L && newuser[0] != '\0' && (user = strdup(newuser)) == 0L)
		NoMem();
	if(newvn != 0L && newvn[0] != '\0' && (vn = strdup(newvn)) == 0L)
		NoMem();
	if(newts != 0L && newts[0] != '\0' && (ts = strdup(newts)) == 0L)
		NoMem();
	if(newoptions != 0L && newoptions[0] != '\0' && (option = strdup(newoptions)) == 0L)
		NoMem();
	if(newtag != 0L && newtag[0] != '\0' && (tag = strdup(newtag)) == 0L)
		NoMem();
	if(newdate != 0L && newdate[0] != '\0' && (date = strdup(newdate)) == 0L)
		NoMem();
	if(newts_conflict != 0L && newts_conflict[0] != '\0' && (ts_conflict = strdup(newts_conflict)) == 0L)
		NoMem();

	if(newvn != 0L && newvn[0] == '-')
		SetRemoved(true);
}


Tagnode::Tagnode(const char *newfullpath, const char *newuser,
	const char *newvn, const char *newts, const char *newoptions, const char *newtag,
	const char *newdate, const char *newts_conflict)
{
	if((fullpath = strdup(newfullpath)) == 0L)
		NoMem();
	user = newuser;
	alreadyChecked = false;
	tag = 0L;
	date = 0L;
}

Tagnode::~Tagnode()
{
	free(fullpath);
	if(tag != 0L)
		free(tag);
	if(date != 0L)
		free(date);
}

// access fields
const char * Tagnode::operator[](int index)
{
	if (!alreadyChecked)
	{
		// don't check twice
		alreadyChecked = true;

		// look in Tags file
		std::string tagString;
		std::string tagFile;
		char cmd;

		tagFile = fullpath;
		tagFile.append("/");
		tagFile.append(user);
		tagFile.append("/CVS/Tag");

		if (Tag_Open(tagString, tagFile.c_str(), &cmd) && !tagString.empty())
		{
			// figure out which sticky thing we have
			switch (cmd) {
			case 'T':
			case 'N':
				if((tag = strdup(tagString.c_str())) == 0L)
					NoMem();
				break;
			case 'D':
				if((date = strdup(tagString.c_str())) == 0L)
					NoMem();
				break;
			}
		}
	}

	// return the right one
	switch (index)
	{
	case EntnodeData::kTagOnly:
		return tag;
	case EntnodeData::kDateOnly:
		return date;
	default:
		return tag != 0L ? tag : date;
	}
}

EntnodeFolder::EntnodeFolder(const char *newfullpath, const char *newuser, const char *newvn,
	const char *newts, const char *newoptions, const char *newtag,
	const char *newdate, const char *newts_conflict) : EntnodeData(ENT_SUBDIR)
{
	if(newuser != 0L && newuser[0] != '\0' && (user = strdup(newuser)) == 0L)
		NoMem();

	if(newvn != 0L && newvn[0] == '-')
		SetRemoved(true);

	tagnode = new Tagnode(newfullpath, user, newvn, newts, newoptions, newtag, newdate, newts_conflict);
}


const char * EntnodeFolder::operator[](int index) const
{
	switch (index)
	{
	case kTag:
	case kTagOnly:
	case kDateOnly:
		return (*tagnode)[index];
	default:
		return EntnodeData::operator[](index);
	}
}

EntnodeOuterFolder::EntnodeOuterFolder(const char *newfullpath, const char *newuser, const char *newvn,
	const char *newts, const char *newoptions, const char *newtag,
	const char *newdate, const char *newts_conflict) :
	EntnodeFolder(newfullpath, newuser, newvn, newts, newoptions, newtag,
		newdate, newts_conflict)
{
}

/* Return the next real Entries line.  On end of file, returns NULL.
   On error, prints an error message and returns NULL.  */

static EntnodeData * fgetentent(FILE *fpin, const char *fullpath, char *cmd = 0L)
{
	EntnodeData *ent;
	char *line = 0L;
	size_t line_chars_allocated = 0;
	register char *cp;
	ent_type type;
	char *l, *user, *vn, *ts, *options;
	char *tag_or_date, *tag, *date, *ts_conflict;
	int line_length;

	ent = 0L;
	while ((line_length = getline (&line, &line_chars_allocated, fpin)) > 0)
	{
		l = line;

		/* If CMD is not NULL, we are reading an Entries.Log file.
		Each line in the Entries.Log file starts with a single
		character command followed by a space.  For backward
		compatibility, the absence of a space indicates an add
		command.  */
		if (cmd != 0L)
		{
			if (l[1] != ' ')
				*cmd = 'A';
			else
			{
				*cmd = l[0];
				l += 2;
			}
		}
		type = ENT_FILE;

		if (l[0] == 'D')
		{
			type = ENT_SUBDIR;
			++l;
			/* An empty D line is permitted; it is a signal that this
			Entries file lists all known subdirectories.  */
		}

		if (l[0] != '/')
			continue;

		user = l + 1;
		if ((cp = strchr (user, '/')) == 0L)
			continue;
		*cp++ = '\0';
		vn = cp;
		if ((cp = strchr (vn, '/')) == 0L)
			continue;
		*cp++ = '\0';
		ts = cp;
		if ((cp = strchr (ts, '/')) == 0L)
			continue;
		*cp++ = '\0';
		options = cp;
		if ((cp = strchr (options, '/')) == 0L)
			continue;
		*cp++ = '\0';
		tag_or_date = cp;
		if ((cp = strchr (tag_or_date, '\n')) == 0L)
			continue;
		*cp = '\0';
		tag = (char *) 0L;
		date = (char *) 0L;
		if (*tag_or_date == 'T')
			tag = tag_or_date + 1;
		else if (*tag_or_date == 'D')
			date = tag_or_date + 1;

		if ((ts_conflict = strchr (ts, '+')))
			*ts_conflict++ = '\0';

		if(type == ENT_SUBDIR)
			ent = new EntnodeFolder(fullpath, user);
		else
			ent = new EntnodeFile(fullpath, user, vn, ts, options,
				tag, date, ts_conflict);
		if(ent == 0L)
			errno = ENOMEM;
		break;
	}

	if (line_length < 0 && !feof (fpin))
	{
		_ASSERT(false);
		// cvs_err("Cannot read entries file (error %d)\n", errno);
	}

	free (line);
	return ent;
}

/* Read the entries file into a list, hashing on the file name.

   UPDATE_DIR is the name of the current directory, for use in error
   messages, or NULL if not known (that is, noone has gotten around
   to updating the caller to pass in the information).  */
bool Entries_Open (std::map<std::string, ENTNODE> & entries, const char *fullpath)
{
	chdir(fullpath);
	chdir("CVS");

	EntnodeData *ent;

	entries.clear();
	FILE *fpin = fopen("Entries", "r");
	if (fpin == 0L)
	{
		return false;
	}

	while ((ent = fgetentent (fpin, fullpath)) != 0L)
	{
		ENTNODE newnode(ent);
		ent->UnRef();

		std::string name = (*newnode.Data())[EntnodeData::kName];
		std::map<std::string, ENTNODE>::iterator it = entries.find(name);
		if (it != entries.end())
		{
			_ASSERT(false);
			//cvs_err("Warning : duplicated entry in the 'CVS/Entries' file in directory '%s'\n",
			//	fullpath);
		}
		entries[name] = newnode;
	}

	fclose (fpin);

	fpin = fopen("Entries.log", "r");
	if (fpin != 0L)
	{
		char cmd;

		while ((ent = fgetentent (fpin, fullpath, &cmd)) != 0L)
		{
			ENTNODE newnode(ent);
			ent->UnRef();

			std::string name = (*newnode.Data())[EntnodeData::kName];

			switch (cmd)
			{
			case 'A':
				entries[name] = newnode;
				break;
			case 'R':
				entries.erase(std::string(name));
				break;
			default:
				/* Ignore unrecognized commands.  */
				break;
			}
		}
		fclose (fpin);
	}
	return true;
}

// mostly the cvs one...
static bool unmodified(const struct stat & sb, const char *ts)
{
    char *cp;
	struct tm *tm_p;
	struct tm local_tm;
	/* We want to use the same timestamp format as is stored in the
	st_mtime.  For unix (and NT I think) this *must* be universal
	time (UT), so that files don't appear to be modified merely
	because the timezone has changed.  For VMS, or hopefully other
	systems where gmtime returns NULL, the modification time is
	stored in local time, and therefore it is not possible to cause
	st_mtime to be out of sync by changing the timezone.  */
	tm_p = gmtime (&sb.st_mtime);
	if (tm_p)
	{
		memcpy (&local_tm, tm_p, sizeof (local_tm));
		cp = asctime (&local_tm);	/* copy in the modify time */
	}
	else
		cp = ctime(&sb.st_mtime);

	cp[24] = 0;

#ifdef _MSC_VER
	/* Work around non-standard asctime() and ctime() in MS VC++.
	These return '01' instead of ' 1' for the day of the month. */
	if (
		(strlen(ts) > 8) &&
		((cp[8] == '0' && ts[8] == ' ') ||
		(cp[8] == ' ' && ts[8] == '0'))
	) {
		cp[8] = ts[8];
	}
#endif

	return strcmp(cp, ts) == 0;
}

EntnodeData *Entries_SetVisited(const char *path, std::map<std::string, ENTNODE> & entries, const char *name,
	const struct stat & finfo, bool isFolder/*, const vector<CStr> * ignlist*/)
{
	bool isCvs = false;
	std::string lookupName;
	if(isFolder)
	{
		EntnodeFolder *adata = new EntnodeFolder(path, name);
		ENTNODE anode(adata);
		adata->UnRef();

		lookupName = (*anode.Data())[EntnodeData::kName];
		std::map<std::string, ENTNODE>::iterator it = entries.find(lookupName);
		isCvs = it != entries.end();
		if (!isCvs)
			entries[lookupName] = anode;
	}
	else
	{
		EntnodeFile *adata = new EntnodeFile(path, name);
		ENTNODE anode(adata);
		adata->UnRef();

		lookupName = (*anode.Data())[EntnodeData::kName];
		std::map<std::string, ENTNODE>::iterator it = entries.find(lookupName);
		isCvs = it != entries.end();
		if (!isCvs)
			entries[lookupName] = anode;
	}

	const ENTNODE & theNode = entries[lookupName];
	EntnodeData *data = ((ENTNODE *)&theNode)->Data();
	data->SetVisited(true);
	if(!isCvs)
	{
		data->SetUnknown(true);
/*		if(ignlist != 0L && MatchIgnoredList(name, *ignlist))
			data->SetIgnored(true); */

		// the folder may have some cvs informations in it, despite the fact
		// that it is not referenced by the parent directory, so try
		// to figure it.

		if(!data->IsIgnored())
		{
			std::string cvsFile = path;
			EnsureTrailingDelimiter(cvsFile);
			cvsFile += name;
			EnsureTrailingDelimiter(cvsFile);
			cvsFile += "CVS";
			struct stat sb;
			if (stat(cvsFile.c_str(), &sb) != -1 && S_ISDIR(sb.st_mode))
			{
				data->SetUnknown(false);
			}
		}
	}

	if(isFolder)
	{
		if(data->IsIgnored())
		{
			data->SetDesc("Ignored Folder");
		}
		else if(data->IsUnknown())
		{
			data->SetDesc("NonCvs Folder");
		}
		else
		{
			data->SetDesc("Folder");
		}
	}
	else
	{
		const char *ts = (*data)[EntnodeFile::kTS];
		if(ts == 0L)
			data->SetUnmodified(true);
		else
			data->SetUnmodified(unmodified(finfo, ts));

		const char *ts_conflict = (*data)[EntnodeFile::kConflict];
		if (ts_conflict == 0)
			data->SetNeedsMerge(false);
		else
			data->SetNeedsMerge(unmodified(finfo,ts_conflict));

		data->SetLocked((finfo.st_mode & S_IWRITE) == 0);
		
		const char *info = 0L;
		if(data->IsIgnored())
		{
			data->SetDesc("Ignored");
		}
		else if(data->IsUnknown())
		{
			data->SetDesc("NonCvs file");
		}
		else if(data->NeedsMerge())
		{
			data->SetDesc("Conflict");
		}
		else if((info = (*data)[EntnodeFile::kOption]) != 0L && strcmp(info, "-kb") == 0)
		{
			data->SetDesc(data->IsUnmodified() ? "Binary" : "Mod. Binary");
		}
		else
		{
			data->SetDesc(data->IsUnmodified() ? "File" : "Mod. File");
		}
	}

	return data;
}

bool Tag_Open(std::string & tag, const char *fullpath, char *cmd)
{
	tag = "";
	if (cmd != 0L)
		*cmd = '\0';

	FILE *fpin = fopen(fullpath, "r");
	if (fpin == 0L)
		return false;

	char *line = 0L;
	size_t line_chars_allocated = 0;
	int line_length;
	if((line_length = getline (&line, &line_chars_allocated, fpin)) > 0)
	{
		char *tmp = strchr(line, '\n');
		if(tmp != 0L)
			*tmp = '\0';
		if (cmd != 0L)
		{
			tag = line + 1;
			*cmd = line[0];
		}
		else {
			if(line[0] == 'T')
				tag = line + 1;
		}
	}
	if(line != 0L)
		free(line);
	if (line_length < 0 && !feof (fpin))
	{
		fclose (fpin);
		_ASSERT(false);
//		cvs_err("Cannot open Tag file in '%s' (error %d)\n", fullpath, errno);
		return false;
	}

	fclose (fpin);
	return !tag.empty();
}

