/*
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 1, or (at your option)
** any later version.

** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.

** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * Author : Alexandre Parenteau <aubonbeurre@hotmail.com> --- April 1998
 */

/*
 * CvsEntries.h --- adaptation from cvs/src/entries.c
 */

#ifndef CVSENTRIES_H
#define CVSENTRIES_H

#ifdef _MSC_VER
#pragma warning(disable:4786 4503)
#endif

#include <map>
#include <string>
// #include "CvsIgnore.h"

typedef enum
{
	ENT_FILE,
	ENT_SUBDIR
} ent_type;

class EntnodeData
{
public:
	EntnodeData(ent_type atype) : ref(1), type(atype), user(0L), desc(0L),
		missing(false), visited(false), unknown(false),
		unmodified(true), needsMerge(false),
		ignored(false), locked(false), removed(false)
	{
#ifdef macintosh
		macspec.vRefNum = 0;
		macspec.parID = 0;
		macspec.name[0] = '\0';
#endif /* macintosh */
	}
	virtual ~EntnodeData();

	enum
	{
		kName = 0,
		kStatus = 3,
		kVN = 1,
		kTS = 5,
		kOption = 2,
		kTag = 4,
		kConflict = 6,
		kTagOnly = 7,
		kDateOnly = 8
	};

	inline ent_type GetType(void) const {return type;}
	inline static int Compare(const EntnodeData & node1, const EntnodeData & node2)
	{
#ifdef WIN32
		return stricmp(node1.user, node2.user);
#else /* !WIN32 */
		return strcmp(node1.user, node2.user);
#endif /* !WIN32 */
	}
	inline EntnodeData *Ref(void) {++ref; return this;}
	inline EntnodeData *UnRef(void)
	{
		if(--ref == 0)
		{
			delete this;
			return 0L;
		}
		return this;
	}

	virtual const char *operator[](int index) const
	{
		if(index == kName)
			return user;
		else if(index == kStatus)
			return desc;
		return 0L;
	}

	inline void SetMissing(bool state) {missing = state;}
	inline bool IsMissing(void) const {return missing;}
	inline void SetVisited(bool state) {visited = state;}
	inline bool IsVisited(void) const {return visited;}
	inline void SetUnknown(bool state) {unknown = state;}
	inline bool IsUnknown(void) const {return unknown;}
	inline void SetIgnored(bool state) {ignored = state;}
	inline bool IsIgnored(void) const {return ignored;}
	inline void SetLocked(bool state) {locked = state;}
	inline bool IsLocked(void) const {return locked;}
	inline void SetRemoved(bool state) {removed = state;}
	inline bool IsRemoved(void) const {return removed;}
	void SetDesc(const char *newdesc);
	inline const char *GetDesc(void) const {return desc;}
	inline void SetUnmodified(bool state) {unmodified = state;}
	inline bool IsUnmodified(void) const {return unmodified;}
	inline void SetNeedsMerge(bool state) {needsMerge = state;}
	inline bool NeedsMerge(void) const {return needsMerge;}

#ifdef macintosh
	FSSpec & MacSpec(void) {return macspec;}
	const FSSpec & GetMacSpec(void) const {return macspec;}
#endif /* macintosh */
protected:
	int ref;
	ent_type type;
	char *user;
	const char *desc;
	bool missing;
	bool visited;
	bool unknown;
	bool unmodified;
	bool needsMerge;
	bool ignored;
	bool locked;
	bool removed;
#ifdef macintosh
	FSSpec macspec;
#endif /* macintosh */
};

class EntnodeFile : public EntnodeData
{
public:
	EntnodeFile(const char *newfullpath, const char *newuser, const char *newvn = 0L,
		const char *newts = 0L, const char *newoptions = 0L,
		const char *newtag = 0L, const char *newdate = 0L,
		const char *newts_conflict = 0L);

	virtual ~EntnodeFile();

	// access fields
	virtual const char *operator[](int index) const
	{
		if(index == kVN)
			return vn;
		else if(index == kTS)
			return ts;
		else if(index == kOption)
			return option;
		else if(index == kTag)
			return tag != 0L ? tag : date;
		else if(index == kConflict)
			return ts_conflict;
		else if(index == kTagOnly)
			return tag;
		else if(index == kDateOnly)
			return date;

		return EntnodeData::operator[](index);
	}

protected:
	char *vn;
	char *ts;
	char *option;
	char *tag; // can be nil
	char *date; // can be nil
	char *ts_conflict; // can be nil
};

class Tagnode
{
public:
	Tagnode(const char *newfullpath, const char *newuser, const char *newvn = 0L,
		const char *newts = 0L, const char *newoptions = 0L, const char *newtag = 0L,
		const char *newdate = 0L, const char *newts_conflict = 0L);

	virtual ~Tagnode();

	// access fields
	virtual const char *operator[](int index);

protected:
	const char *user;
	char *fullpath;

	bool alreadyChecked;
	char *tag; // from CVS/Tag - can be nil
	char *date; // from CVS/Tag - can be nil
};

class EntnodeFolder : public EntnodeData
{
public:
	EntnodeFolder(const char *newfullpath, const char *newuser, const char *newvn = 0L,
		const char *newts = 0L, const char *newoptions = 0L, const char *newtag = 0L,
		const char *newdate = 0L, const char *newts_conflict = 0L);

	virtual ~EntnodeFolder()
	{
		delete tagnode;
	}

	// access fields
	virtual const char *operator[](int index) const;

protected:
	Tagnode * tagnode; // for sticky tag info
};

class EntnodeOuterFolder : public EntnodeFolder
{
public:
	EntnodeOuterFolder(const char *newfullpath, const char *newuser, const char *newvn = 0L,
		const char *newts = 0L, const char *newoptions = 0L, const char *newtag = 0L,
		const char *newdate = 0L, const char *newts_conflict = 0L);
};

// ENTNODE gets a node file or a node folder
// and reference it.
class ENTNODE
{
public:
	ENTNODE() : shareData(0L) {}
	ENTNODE(EntnodeData *data) : shareData(data->Ref()) {}
	ENTNODE(EntnodeData & data) : shareData(data.Ref()) {}
	ENTNODE(const ENTNODE & anode) : shareData(0L)
	{
		*this = anode;
	}
	~ENTNODE()
	{
		if(shareData != 0L)
			shareData->UnRef();
	}

	inline static int Compare(const ENTNODE & node1, const ENTNODE & node2)
	{
		return EntnodeData::Compare(*node1.shareData, *node2.shareData);
	}

	inline ENTNODE & operator=(const ENTNODE & anode)
	{
		if(shareData != 0L)
		{
			shareData->UnRef();
			shareData = 0L;
		}
		if(anode.shareData != 0L)
			shareData = ((ENTNODE *)&anode)->shareData->Ref();
		return *this;
	}

	inline EntnodeData *Data(void) {return shareData;}
protected:
	EntnodeData *shareData;
};

bool Entries_Open (std::map<std::string, ENTNODE> & entries, const char *fullpath);
	// return false if no CVS/Entries

EntnodeData *Entries_SetVisited(const char *path, std::map<std::string, ENTNODE> & entries, const char *name,
	const struct stat & finfo, bool isFolder/*, const vector<CStr> * ignlist = 0L*/);
	// fill an ENTNODE when the file appears on the disk : it set
	// some flags like "visited", "unknown", "ignored"... and return
	// a reference to the node.

bool Tag_Open(std::string & tag, const char *fullpath, char *cmd = 0L);
	// return false if no CVS/Tag

#endif /* CVSENTRIES_H */
