// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - September 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifdef _MSC_VER
// turn off symbol-length warnings
#pragma warning(disable:4786 4503)
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "DefaultMenus.h"
#include "HasMenu.h"
#include "../Utils/CrapLexer.h"
#include "../Utils/TortoiseUtils.h"
#include <algorithm>    // for transform
#ifdef __GNUC__
	#include <ctype.h>  // for tolower
#else
	#include <locale>   // for tolower
#endif
#include <map>

static bool SlurpFlags( CrapLexer& lex, int& flags );
static void ShowLexErr( CrapLexer& lex, std::string const& msg = "Syntax error" );

void FillInDefaultMenus(std::vector<MenuDescription>& menus)
{
	CrapLexer lex( GetTortoiseFolder()+"TortoiseMenus.config" );

	bool done = false;
	while( !done )
	{
		int type;
		std::string tok;

		type = lex.GetTok( tok );
		if( type == CrapLexer::typeFinished )
			break;
		else if( type == CrapLexer::typeErr )
		{
			ShowLexErr( lex, "Unexpected end of file" );
			// message here!
			return;
		}
		else
		{
			std::transform( tok.begin(), tok.end(), tok.begin(), tolower );
			if( tok == "item" )
			{
				std::string verb;
				std::string helptext;
				std::string menutext;
				int flags;

				if( lex.GetTok( verb ) != CrapLexer::typeString )
				{
					ShowLexErr( lex );
					return;
				}
				if( lex.GetTok( menutext ) != CrapLexer::typeString )
				{
					ShowLexErr( lex );
					return;
				}
				if( !SlurpFlags( lex, flags ) )
				{
					ShowLexErr( lex );
					return;
				}
				if( lex.GetTok( helptext ) != CrapLexer::typeString )
				{
					ShowLexErr( lex );
					return;
				}

				menus.push_back(
					MenuDescription( verb, menutext, flags, helptext ) );
			}
			else if( tok == "separator" )
			{
				int flags;
				if( !SlurpFlags( lex, flags ) )
				{
					ShowLexErr( lex );
					return;
				}
				menus.push_back(
					MenuDescription( "", "", flags, "" ) );
			}
			else
			{
				ShowLexErr( lex );
				return;
				//error
			}

		}
	}
}



static bool SlurpFlags( CrapLexer& lex, int& flags )
{
	static bool init=false;
	static std::map< std::string, int > flagtable;
	if( !init )
	{
		flagtable[ "on_sub_menu" ] = ON_SUB_MENU;
		flagtable[ "exactly_one" ] = EXACTLY_ONE;
		flagtable[ "always_has" ] = ALWAYS_HAS;
		flagtable[ "all_must_match_type" ] = ALL_MUST_MATCH_TYPE;
		flagtable[ "type_not_in_cvs" ] = TYPE_NOTINCVS;
		flagtable[ "type_nowhere_near_cvs" ] = TYPE_NOWHERENEAR_CVS;
		flagtable[ "type_in_cvs" ] = TYPE_INCVS;
		flagtable[ "type_changed" ] = TYPE_CHANGED;
		flagtable[ "type_conflict" ] = TYPE_CONFLICT;
		flagtable[ "type_outer_dir" ] = TYPE_OUTERDIRECTORY;
		flagtable[ "type_directory" ] = TYPE_DIRECTORY;
		flagtable[ "type_file" ] = TYPE_FILE;
		flagtable[ "type_removeable" ] = TYPE_REMOVEABLE;
		init = true;
	}

	flags = 0;
	std::string tok;
	if( lex.GetTok( tok ) != CrapLexer::typeString )
		return false;
	if( tok != "(" )
		return false;

	while( true )
	{
		if( lex.GetTok( tok ) != CrapLexer::typeString )
			return false;

		std::transform( tok.begin(), tok.end(), tok.begin(), tolower );

		if( tok == ")" )
			break;	// done.

		std::map< std::string, int >::const_iterator it =
			flagtable.find( tok );

		if( it == flagtable.end() )
			return false;

		flags |= it->second;
	}
	return true;
}




static void ShowLexErr( CrapLexer& lex, std::string const& msg )
{
	char buf[18];
	char *lineNum = itoa(lex.LineNum(), buf, 10);

	std::string fmt;
	fmt = "Error while parsing config file\r\n";
	fmt += "(\"" + lex.Filename() + "\" line ";
	fmt += std::string(lineNum) + "):\r\n";
	fmt + msg;

	::MessageBox(NULL, fmt.c_str(), "TortoiseShell.dll", MB_OK );
	return;
}

