// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Ben Campbell
// <ben.campbell@creaturelabs.com> - September 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifdef _MSC_VER
#pragma warning(disable:4786 4503)
#endif

#include "CommitDialog.h"
#include "WxwHelpers.h"
#include "../Utils/PathUtils.h"
#include "../Utils/TortoiseUtils.h"
#include "../Utils/ShellUtils.h"

//static
bool DoCommitDialog(const std::vector<std::string>& modified,
		const std::vector<std::string>& added, const std::vector<std::string>& removed,
		std::string& comment, std::vector<std::string>& userSelection)
{
	CommitDialog* dlg = new CommitDialog(modified, added, removed, comment);
	bool ret = (dlg->ShowModal() == wxID_OK);

	if (ret)
	{
		comment = dlg->myComment->GetValue();

		// Fill in user selection of ticked items
		for(unsigned int i = 0; i < modified.size(); ++i)
			if (dlg->myModified->IsChecked(i))
				userSelection.push_back(modified[i]);
		for(unsigned int j = 0; j < added.size(); ++j)
			if (dlg->myAdded->IsChecked(j))
				userSelection.push_back(added[j]);
		for(unsigned int k = 0; k < removed.size(); ++k)
			if (dlg->myRemoved->IsChecked(k))
				userSelection.push_back(removed[k]);
	}

	DestroyDialog(dlg);
	return ret;
}

// Strip shared bits from the front of all the paths
void CommitDialog::ShortenPaths(std::vector<std::string>& files, std::string& stub)
{
	if (files.empty())
		return;

	while (true)
	{
		std::string firstPart;
		for(unsigned int i = 0; i < files.size(); ++i)
		{
			std::string nextFirstPart = GetFirstPart(files[i]);
			if (nextFirstPart.empty())
				return;

			// If a first chunk of the path doesn't match, we 
			// give up stripping any more
			if (firstPart != "" && firstPart != nextFirstPart)
				return;
			firstPart = nextFirstPart;

			// And we need something left as well!
			if (StripFirstPart(files[i]).empty())
				return;
		}

		// Remember the stub
		stub += GetFirstPart(files[0]);

		// Right, we now know we can strip the front off everything
		for(unsigned int j = 0; j < files.size(); ++j)
		{
			files[j] = StripFirstPart(files[j]);
		}
	}
}

CommitDialog::CommitDialog(std::vector<std::string> modified, std::vector<std::string> added, std::vector<std::string> removed, const std::string& defaultComment)
	: wxDialog(NULL, -1, wxString("TortoiseCVS - Commit"),
	wxDefaultPosition, wxDefaultSize, wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER)
{
	SetIcon(wxIcon("TORTOISE", wxBITMAP_TYPE_ICO_RESOURCE));

	// Trim paths to sensible length
	ShortenPaths(modified, myModifiedStub);
	ShortenPaths(added, myAddedStub);
	ShortenPaths(removed, myRemovedStub);

	// Comment
	wxStaticText *label0 = new wxStaticText(this, -1, "Comment:");
	myComment = new wxTextCtrl(this, -1, defaultComment.c_str(), wxDefaultPosition, wxDefaultSize, wxTE_MULTILINE);
	myComment->SetFocus();
	myComment->SetSelection(0, defaultComment.size());

	// All the file lists horizontally together
	wxBoxSizer *sizerHoriz = new wxBoxSizer(wxHORIZONTAL);

	// Modified
	if (!modified.empty())
	{
		wxStaticText *label1 = new wxStaticText(this, -1, wxString("Modified in ") + myModifiedStub.c_str());
		std::vector<wxString> modifiedWx = ConvertToWxStringArray(modified);
		myModified = new wxCheckListBox(this, -1, wxDefaultPosition, wxDefaultSize, modifiedWx.size(), &modifiedWx[0], 0);
		for(unsigned int i = 0; i < modifiedWx.size(); ++i)
			myModified->Check(i, true);
		wxBoxSizer *sizer1 = new wxBoxSizer(wxVERTICAL);
		sizer1->Add(label1, 0, wxGROW | wxALL, 3);
		sizer1->Add(myModified, 1, wxGROW | wxALL, 3);
		myModified->Connect(-1, wxEVT_RIGHT_DOWN, reinterpret_cast<wxObjectEventFunction>(&CommitDialog::ContextMenuModified));
		sizerHoriz->Add(sizer1, 1, wxGROW | wxALL, 3);
	}

	// Added
	if (!added.empty())
	{
		wxStaticText *label2 = new wxStaticText(this, -1, wxString("Added to ") + myAddedStub.c_str());
		std::vector<wxString> addedWx = ConvertToWxStringArray(added);
		myAdded = new wxCheckListBox(this, -1, wxDefaultPosition, wxDefaultSize, addedWx.size(), &addedWx[0], 0);
		for(unsigned int j = 0; j < addedWx.size(); ++j)
			myAdded->Check(j, true);
		wxBoxSizer *sizer2 = new wxBoxSizer(wxVERTICAL);
		sizer2->Add(label2, 0, wxGROW | wxALL, 3);
		sizer2->Add(myAdded, 1, wxGROW | wxALL, 3);
		myAdded->Connect(-1, wxEVT_RIGHT_DOWN, reinterpret_cast<wxObjectEventFunction>(&CommitDialog::ContextMenuAdded));
		sizerHoriz->Add(sizer2, 1, wxGROW | wxALL, 3);
	}

	// Removed
	if (!removed.empty())
	{
		wxStaticText *label3 = new wxStaticText(this, -1, wxString("Removed from ") + myRemovedStub.c_str());
		std::vector<wxString> removedWx = ConvertToWxStringArray(removed);
		myRemoved = new wxCheckListBox(this, -1, wxDefaultPosition, wxDefaultSize, removedWx.size(), &removedWx[0], 0);
		for(unsigned int k = 0; k < removedWx.size(); ++k)
			myRemoved->Check(k, true);
		wxBoxSizer *sizer3 = new wxBoxSizer(wxVERTICAL);
		sizer3->Add(label3, 0, wxGROW | wxALL, 3);
		sizer3->Add(myRemoved, 1, wxGROW | wxALL, 3);
		myRemoved->Connect(-1, wxEVT_RIGHT_DOWN, reinterpret_cast<wxObjectEventFunction>(&CommitDialog::ContextMenuRemoved));
		sizerHoriz->Add(sizer3, 1, wxGROW | wxALL, 3);
	}

	// OK/Cancel button
	wxBoxSizer *sizerConfirm = new wxBoxSizer(wxHORIZONTAL);
    myOK = new wxButton(this, wxID_OK, "OK (without comment)	");
	myOK->SetDefault();
    wxButton* cancel = new wxButton(this, wxID_CANCEL, "Cancel");
	sizerConfirm->Add(myOK, 0, wxGROW | wxALL, 5);
    sizerConfirm->Add(cancel, 0, wxGROW | wxALL, 5);

	// Main box with everything in it
	wxBoxSizer *sizerTop = new wxBoxSizer(wxVERTICAL);
	sizerTop->Add(label0, 0, wxGROW | wxALL, 3);
	sizerTop->Add(myComment, 1, wxGROW | wxALL, 3);
	sizerTop->Add(sizerHoriz, 2, wxGROW | wxALL, 3);
	sizerTop->Add(sizerConfirm, 0, wxALIGN_CENTER | wxLEFT | wxRIGHT, 10);

	// Wire up context menus
	Connect(CD_WEB_LOG_REMOVED, wxEVT_COMMAND_MENU_SELECTED, reinterpret_cast<wxObjectEventFunction>(&CommitDialog::WebLogRemoved));
	Connect(myComment->GetId(), wxEVT_COMMAND_TEXT_UPDATED, reinterpret_cast<wxObjectEventFunction>(&CommitDialog::EmptyCommentWarning));
	EmptyCommentWarning();

	// Overall dialog layout settings
    SetAutoLayout(TRUE);
    SetSizer(sizerTop);
    sizerTop->SetSizeHints(this);
    sizerTop->Fit(this);
}

// static
void CommitDialog::ContextMenuAdded(const wxMouseEvent& event)
{
	// TODO: Dedodgify. This is actually called with this being the list box,
	// so we find CommitDialog via the parent.
	wxWindow* parent = GetParent();
	CommitDialog* commitDlg = (CommitDialog*) parent;
	std::string file = std::string(commitDlg->myAddedStub) + commitDlg->myAdded->GetStringSelection().c_str();
	POINT p;
	event.GetPosition(&p.x, &p.y);
	p.x += GetPosition().x;
	p.y += GetPosition().y;
	DoExplorerMenu((HWND)(commitDlg->GetHandle()), file.c_str(), p);
	return;
}

// static
void CommitDialog::ContextMenuModified(const wxMouseEvent& event)
{
	// TODO: Dedodgify. This is actually called with this being the list box,
	// so we find CommitDialog via the parent.
	wxWindow* parent = GetParent();
	CommitDialog* commitDlg = (CommitDialog*) parent;
	std::string file = std::string(commitDlg->myModifiedStub) + commitDlg->myModified->GetStringSelection().c_str();
	POINT p;
	event.GetPosition(&p.x, &p.y);
	p.x += GetPosition().x;
	p.y += GetPosition().y;
	DoExplorerMenu((HWND)(commitDlg->GetHandle()), file.c_str(), p);
	return;
}

// static
void CommitDialog::ContextMenuRemoved(const wxMouseEvent& event)
{
    wxMenu menu;
    menu.Append(CD_WEB_LOG_REMOVED, "CVS &Web Log");
	PopupMenu(&menu, event.GetPosition());
}

void CommitDialog::WebLogRemoved()
{
	std::string file = std::string(myRemovedStub) + myRemoved->GetStringSelection().c_str();
	LaunchTortoiseAct("CVSWebLog", file);
}

// Display warning label if they don't enter a comment
void CommitDialog::EmptyCommentWarning()
{
	wxString value = myComment->GetValue();
	if (value.empty())
		myOK->SetLabel("OK (without comment)");
	else
		myOK->SetLabel("OK");
}

long CommitDialog::MSWWindowProc(WXUINT nMsg, WXWPARAM wParam, WXLPARAM lParam)
{
	MSG msg;
	msg.message = nMsg;
	msg.wParam = wParam;
	msg.lParam = lParam;

	// Pass owner draw on to Explorer popup menu
	if (!DoExplorerMenuWndProc(msg))
	{
		// Otherwise carry on ourself
		return wxDialog::MSWWindowProc(nMsg, wParam, lParam);
	}

	return 0;
}

