// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - October 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "PreferencesDialog.h"
#include <wx/notebook.h>
#include "WxwHelpers.h"
#include "../Utils/Preference.h"
#include "../Utils/TortoiseDebug.h"

// Helper class used to store an int in wxWindows events
class IndexData : public wxObject
{
public:
	IndexData(int data) : myData(data) {};
	int myData;
};

//static
void DoPreferencesDialog(TortoisePreferences& prefs)
{
	PreferencesDialog* dlg = new PreferencesDialog(prefs);
	bool ret = dlg->ShowModal() == wxID_OK;

	if (ret)
	{
		for (unsigned int i = 0; i < dlg->myPrefs.GetCount(); ++i)
			dlg->ExtractPreference(dlg->myPrefs.Lookup(i), dlg->myWidgets[i]);
	}

	DestroyDialog(dlg);
}

PreferencesDialog::PreferencesDialog(TortoisePreferences& prefs)
	: wxDialog(NULL, -1, wxString("TortoiseCVS Preferences"),
	wxDefaultPosition, wxDefaultSize, wxRESIZE_BORDER | wxDEFAULT_DIALOG_STYLE),
	myPrefs(prefs)

{
	SetIcon(wxIcon("TORTOISE", wxBITMAP_TYPE_ICO_RESOURCE));

	// OK/Cancel button
	wxBoxSizer *sizerConfirm = new wxBoxSizer(wxHORIZONTAL);
    wxButton* ok = new wxButton(this, wxID_OK, "OK");
	ok->SetDefault();
    wxButton* cancel = new wxButton(this, wxID_CANCEL, "Cancel");
    sizerConfirm->Add(ok, 0, wxGROW | wxALL, 5); 
    sizerConfirm->Add(cancel, 0, wxGROW | wxALL, 5);

	// Make a notebook, and add preferences to pages
	myNotebook = new wxNotebook(this, -1);
	wxNotebookSizer *nbs = new wxNotebookSizer(myNotebook);
	myWidgets.resize(myPrefs.GetCount());
	myWidgetsExtra.resize(myPrefs.GetCount());
	for (unsigned int i = 0; i < myPrefs.GetCount(); ++i)
		AddPreference(myPrefs.Lookup(i), myWidgets[i], myWidgetsExtra[i], i);

	// Add all the pages to the notebook
	std::map<std::string, wxPanel*>::iterator it;
	for (it = myPages.begin(); it != myPages.end(); it++)
		myNotebook->AddPage(it->second, it->first.c_str());

	// Main vertical with controls and buttons in it
	wxBoxSizer *sizerTop = new wxBoxSizer(wxVERTICAL);
	sizerTop->Add(nbs, 1, wxGROW | wxALL, 8);
	sizerTop->Add(sizerConfirm, 0, wxALIGN_CENTER | wxALL, 0);

	// Overall dialog layout settings
    SetAutoLayout(TRUE);
    SetSizer(sizerTop);
    sizerTop->SetSizeHints(this);
    sizerTop->Fit(this);
	RefreshGhosting();
}

void PreferencesDialog::AddPreference(const Preference& pref, wxControl* &widget, wxControl* &widgetExtra, int ix)
{
	// See if page already exists
	std::string pageName = pref.myPage;
	std::map<std::string, wxPanel*>::iterator it = myPages.find(pageName);
	if (it == myPages.end())
	{
		// Page doesn't exist, so add it in
		wxPanel *page = new wxPanel(myNotebook);
		wxFlexGridSizer *sizer = new wxFlexGridSizer(2, 5, 5);
		page->SetAutoLayout(TRUE);
		page->SetSizer(sizer);
		sizer->AddGrowableCol(1);
		myPages[pageName] = page;
	}
	// We have the sizer for this page
	wxPanel *page = myPages.find(pageName)->second;
	wxSizer *gridSizer = page->GetSizer();

	wxStaticText* label = new wxStaticText(page, -1, ((pref.myDescription + ":").c_str()));
	gridSizer->Add(label, 1, wxSHAPED | wxALIGN_CENTER_VERTICAL, 5);
	widgetExtra = NULL;
	
	switch (pref.myType)
	{
		case Preference::STRING:
		case Preference::FILEPICK: // TODO, like DIRPICK below
		{
			widget = new wxTextCtrl(page, -1, pref.myStringData.c_str());
			gridSizer->Add(widget, 1, wxGROW | wxALL, 5);
			break;
		}
		case Preference::DIRPICK:
		{
			wxBoxSizer *sizer = new wxBoxSizer(wxHORIZONTAL);
			widget = new wxTextCtrl(page, -1, pref.myStringData.c_str());
			wxButton *button = new wxButton(page, -1, "Browse...");
			sizer->Add(widget, 2, wxEXPAND | wxLEFT | wxRIGHT, 0);
			sizer->Add(button, 0, wxLEFT, 5);
			// We use IndexData because we need a real wxObject because
			// wxWindows deletes it for us.  See TODO below.
			Connect(button->GetId(),
                    wxEVT_COMMAND_BUTTON_CLICKED,
                    reinterpret_cast<wxObjectEventFunction>(&PreferencesDialog::DirectoryBrowse),
                    dynamic_cast<wxObject*>(new IndexData(ix)));
	
			gridSizer->Add(sizer, 1, wxGROW | wxALL, 5);
			widgetExtra = button;
			break;
		}
		case Preference::BOOLEAN:
		{
			wxCheckBox* check = new wxCheckBox(page, -1, pref.myCheckBoxText.c_str());
			check->SetValue(pref.myBooleanData);
			Connect(check->GetId(),
                wxEVT_COMMAND_CHECKBOX_CLICKED,
                reinterpret_cast<wxObjectEventFunction>(&PreferencesDialog::RefreshGhosting));
			widget = check;
			gridSizer->Add(widget, 1, wxGROW | wxALL, 5);
			break;
		}
		case Preference::ENUMERATION:
		{
			// Add each possibility in
			std::vector<wxString> choices;
			for (unsigned int i = 0; i < pref.myEnumerationPossibilities.size(); ++i)
				choices.push_back(pref.myEnumerationPossibilities[i].c_str());

			// Make widget
			wxChoice* choice = new wxChoice(page, -1, wxDefaultPosition, wxDefaultSize, choices.size(), &choices[0]);
			widget = choice;

			// Set default
			choice->SetSelection(pref.myEnumerationData);

			gridSizer->Add(widget, 1, wxGROW | wxALL, 5);
			break;

		}
    }

	// Tooltips
	label->SetToolTip(pref.myTooltip.c_str());
	widget->SetToolTip(pref.myTooltip.c_str());
}

void PreferencesDialog::ExtractPreference(Preference& pref, wxControl* widget)
{
	switch (pref.myType)
	{
		case Preference::STRING:
		case Preference::DIRPICK:
		case Preference::FILEPICK:
		{
			pref.myStringData = static_cast<wxTextCtrl*>(widget)->GetValue();
			break;
		}
		case Preference::BOOLEAN:
		{
			pref.myBooleanData = static_cast<wxCheckBox*>(widget)->GetValue();
			break;
		}
		case Preference::ENUMERATION:
		{
			pref.myEnumerationData = static_cast<wxChoice*>(widget)->GetSelection();
			break;
		}
	}
}

void PreferencesDialog::RefreshGhosting()
{
		
	for (unsigned int i = 0; i < myPrefs.GetCount(); ++i)
	{
		Preference& pref = myPrefs.Lookup(i);
		if (pref.myGhostPreference >= 0)
		{
			ASSERT(myPrefs.Lookup(pref.myGhostPreference).myType == Preference::BOOLEAN);
			bool value = static_cast<wxCheckBox*>(myWidgets[pref.myGhostPreference])->GetValue();
			if (!pref.myGhostPositive)
				value = !value;
			myWidgets[i]->Enable(value);	
			if (myWidgetsExtra[i])
				myWidgetsExtra[i]->Enable(value);

		}
	}
}

void PreferencesDialog::DirectoryBrowse(wxEvent& event)
{
	// TODO: Using member m_callbackUserData is undocumented in wxWindows.
	// One of the following should be done:
	// - Find better way of doing this in wxWindows.  Is this the only way of
	// passing user data to an event when there are possibly multiple instances
	// of the same button?
	// - Make current way in wxWindows more explicit.  This would mean adding
	// event.GetUserData() function and documenting it, and cross referencing
	// with the last parameter of wxEvtHandler::Connect().
	int ix = dynamic_cast<IndexData*>(event.m_callbackUserData)->myData;
	
	wxTextCtrl* text = static_cast<wxTextCtrl*>(myWidgets[ix]);
	wxString description = myPrefs.Lookup(ix).myDescription.c_str();

	wxDirDialog* dlg = new wxDirDialog(NULL, description,
		text->GetValue().c_str());

	if (dlg->ShowModal() == wxID_OK)
		text->SetValue(dlg->GetPath());

	DestroyDialog(dlg);
}


