// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - October 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifdef _MSC_VER
#pragma warning(disable:4786 4503)
#endif

#include "WebLog.h"
#include "../Utils/TortoiseRegistry.h"
#include "../Utils/PathUtils.h"
#include "../Utils/Keyboard.h"

#include "../Utils/Thread.h"
#include "ThreadSafeProgress.h"

#include <algorithm>

// static
bool WebLog::CheckURLWorks(const std::string& url)
{
	if (url == "plainlog")
		return true;

	// Using HTTP query the path on the server.  If we get an error (i.e. >=400)
	// then tell the user, and query for a new URL-stub (this covers the location
	// on the server changing).

	URL_COMPONENTS urlComp;
	memset(&urlComp, 0, sizeof(urlComp));
	char hostName[_MAX_PATH];
	char urlPath[_MAX_PATH];
	urlComp.lpszHostName = hostName;
	urlComp.dwHostNameLength = _MAX_PATH - 1;
	urlComp.lpszUrlPath = urlPath;
	urlComp.dwUrlPathLength = _MAX_PATH - 1;
	urlComp.dwStructSize = sizeof(urlComp);

	bool result = true;
	if (!InternetCrackUrl(url.c_str(), 0, ICU_DECODE, &urlComp))
		result = false;
	else
	{
		HttpSnaffle snaffler;
		if (!snaffler.Open("TortoiseCVS URL tester", urlComp.lpszHostName))
			result = false;
		if (!snaffler.StartFetch(urlComp.lpszUrlPath))
			result = false;
		if (snaffler.FetchStatusCode() >= 400)
			result = false;
	}

	return result;
}

// static
std::string WebLog::SourceForgeProject(const std::string& server)
{
	if (server.find(".sourceforge.net") == std::string::npos)
		return "";

	std::string::size_type pos = server.find_first_of(".");
	if (pos == std::string::npos || pos >= server.size() - 1)
		return "";

	std::string endPart = server.substr(pos + 1);
	pos = endPart.find_first_of(".");
	if (pos == std::string::npos)
		return "";

	std::string project = endPart.substr(0, pos);
	std::transform( project.begin(), project.end(), project.begin(), tolower );
	return project;
}

WebLog::WebLog(const std::string& server, const std::string& ambientPart)
: myServer(server), myAmbientPart(ambientPart), 
  myQueriedBefore(false), myThreadSafeProgress(NULL)
{

}


bool WebLog::TrySingleURL(const std::string& urlStubPrefix, const std::string& urlStubSuffix)
{
	std::string url = MakeFullURL(urlStubPrefix, urlStubSuffix);
	std::string message = "Trying " + url + "...\n";
	if (myThreadSafeProgress)
	{
		myThreadSafeProgress->LookBusy();
		myThreadSafeProgress->NewText(const_cast<char*>(message.c_str()), message.size());
	}
	return CheckURLWorks(url);
}

bool WebLog::ScanForURL(std::string& urlStubPrefix, std::string& urlStubSuffix)
{
	const char* firstPart[] =
		{ "/cgi-bin", "/viewcvs", "/cvsweb", "/bin", "" };
	const char* secondPart[] =
		{ "/viewcvs.cgi", "/viewcvs", "/cvsweb.cgi", "/cvsweb.pl", "/cvsweb"};

	// Preprocess to get a list of prefixes to scan
	std::vector<std::string> prefixes;
	std::vector<std::string> suffixes;
	for (unsigned int j = 0; j < sizeof(secondPart) / sizeof(char*); ++j)
	{
		for (unsigned int i = 0; i < sizeof(firstPart) / sizeof(char*); ++i)
		{
			std::string first = firstPart[i];
			std::string second = secondPart[j];
			// Don't mix viewcvs and cvsweb parts, as that is silly and unlikely!
			if (!((first.find("viewcvs") != std::string::npos && second.find("cvsweb") != std::string::npos)
				|| (first.find("cvsweb") != std::string::npos && second.find("viewcvs") != std::string::npos)))
			{
				prefixes.push_back(first + second);
			}
		}
	}

	// Sourceforge special case - add the project name into the path
	std::string sfProject = SourceForgeProject(myServer);
	if (!sfProject.empty())
	{
		for (unsigned int k = 0; k < prefixes.size(); ++k)
			prefixes[k] += "/" + sfProject;
	}

	// Try with no suffix (old Sourceforge used to need a suffix
	// at the end of the URL ?project=cvsgui, for example.  This 
	// isn't the case any more.)
	suffixes.push_back("");

	// Actually perform the scan
	bool result = false;
	for (unsigned int l = 0; l < suffixes.size(); ++l)
	{
		for (unsigned int k = 0; k < prefixes.size(); ++k)
		{
			urlStubPrefix = "http://" + myServer + prefixes[k];
			urlStubSuffix = suffixes[l];
			result = TrySingleURL(urlStubPrefix, urlStubSuffix);
			if (result || myThreadSafeProgress->UserAborted())
				break;
		}
		if (result || myThreadSafeProgress->UserAborted())
			break;
	}

	return result;
}

std::string WebLog::MakeFullURL(const std::string& urlStubPrefix, const std::string& urlStubSuffix)
{
	// Special case, not a real URL (means do plain CVS text log locally)
	if (urlStubPrefix == "plainlog")
		return urlStubPrefix;

	return urlStubPrefix + myAmbientPart + urlStubSuffix;
}

void WebLog::ThreadMain()
{
	std::string message = "Scanning for web log URL on server\n";
	myThreadSafeProgress->NewText(const_cast<char*>(message.c_str()), message.size());

	// Test for obsolete Sourceforge registry keys
	// (it used to use a suffix)
	std::string sfProject = SourceForgeProject(myServer);
	if (!sfProject.empty())
	{
		if (myThreadSuffix != "")
			myThreadPrefix = "scanagain";
	}

	// If there is a URL in the registry, see if it is OK first
	if (!myThreadPrefix.empty() && myThreadPrefix != "scanagain")
		myThreadResult = TrySingleURL(myThreadPrefix, myThreadSuffix);

	if (myThreadSafeProgress->UserAborted())
	{
		myThreadSafeProgress->Finished();
		return;
	}

	// If that doesn't work, look for other values
	if (!myThreadResult)
		myThreadResult = ScanForURL(myThreadPrefix, myThreadSuffix);

	myThreadSafeProgress->Finished();
}

bool WebLog::AutomaticallyFindURL(std::string& urlStubPrefix, std::string& urlStubSuffix)
{
	// Look for an already specified URL-stub for that server in the registry
	urlStubPrefix = TortoiseRegistry::ReadString("WebLogURLPrefix " + myServer);
	urlStubSuffix = TortoiseRegistry::ReadString("WebLogURLSuffix " + myServer);

	// Plain "cvs log" is always OK
	if (urlStubPrefix == "plainlog")
		return true;

	// Make thread to do the scanning, with progress dialog in this thread
	myThreadResult = false;
	ProgressDialogAbstract* progress = CreateProgressDialog();
	myThreadSafeProgress = new ThreadSafeProgress(progress);
	myThreadPrefix = urlStubPrefix;
	myThreadSuffix= urlStubSuffix;
	Thread<WebLog> thread;
	thread.BeginThread(this, &WebLog::ThreadMain);
	myThreadSafeProgress->Main();
	urlStubPrefix = myThreadPrefix;
	urlStubSuffix = myThreadSuffix;
	bool result = myThreadResult;
	// progress->WaitForAbort(); // for testing
	delete myThreadSafeProgress;
	DeleteProgressDialog(progress);
	myThreadSafeProgress = NULL;

	if (result)
	{
		// Store it in registry
		TortoiseRegistry::WriteString("WebLogURLPrefix " + myServer, urlStubPrefix);
		TortoiseRegistry::WriteString("WebLogURLSuffix " + myServer, urlStubSuffix);
	}

	return result;
}

// Query the user for a URL prefix and suffix for his cvsweb/viewcvs-style program
// Results are stored in the registry, so next automatic scan picks them up
bool WebLog::QueryUser(bool queryWasForced, std::string& urlStubPrefix, std::string& urlStubSuffix)
{
	// Work out what text to display
	std::string message;
	if (queryWasForced)
	{
		message += "Please enter the URL where ViewCVS or CVSWeb is. ";
	}
	else
	{
		if (myQueriedBefore)
		{
			message += "Sorry! The URL isn't responding, or is responding with an error.  "
				"This could happen because you are not connected to the internet, because "
				"the server is down, or because the path is wrong.  "
				"If you are behind a firewall which requires a proxy, TortoiseCVS is using "
				"the Internet Explorer proxy.  You might want to set it and try again.\n\n"
				
				"Please either try entering the URL again, or choose the plain text log.  ";
		}
		else
		{
			message += "TortoiseCVS couldn't find the URL on your server where ViewCVS "
				"or CVSWeb is. Please enter the URL manually. ";
		}
	}

	// Explain how they can work it out
	message += 
		"You might be able to find it by browsing on the project's web site for "
		"a web interface to CVS.  Alternatively ask your server administrator.  "
		"Then copy the top level URL into the prefix box below. "
		"Usually you can leave the suffix box blank - it adds extra options to the end of the URL.  "
		// TODO: Detect this case, and do something better:
		"\n\nThis dialog can also appear if you are trying to log "
		"a file that has never been comitted.";

	// As a default - the combo box remembers more that the user has entered
	urlStubPrefix = "http://" + myServer + "/cgi-bin/cvsweb.cgi";
	urlStubSuffix = "";
	bool result = DoLogConfigDialog(urlStubPrefix, urlStubSuffix, message, myServer);
	RemoveTrailingDelimiter(urlStubPrefix);
	if (urlStubPrefix != "scanagain")
		myQueriedBefore = true;
	else
		myQueriedBefore = false;

	// Store what the user did, so the next automatic scan picks it up from the registry
	if (result)
	{
		TortoiseRegistry::WriteString("WebLogURLPrefix " + myServer, urlStubPrefix);
		TortoiseRegistry::WriteString("WebLogURLSuffix " + myServer, urlStubSuffix);
	}

	return result;
}

bool WebLog::PerformSearchAndQuery(std::string& url)
{
	std::string urlStubPrefix, urlStubSuffix;
	bool queryWasForced = false;

	// Holding down control disables the autoscan, and does a query straight away
	if (IsControlDown())
	{
		queryWasForced = true;
	}
	else
	{
		// They're not holding it down, so do the autoscan now
		if (AutomaticallyFindURL(urlStubPrefix, urlStubSuffix))
		{
			url = MakeFullURL(urlStubPrefix, urlStubSuffix);
			return true;
		}
	}

	while (true)
	{
		// Ask the user for a URL, give up if they cancel
		if (!QueryUser(queryWasForced, urlStubPrefix, urlStubSuffix))
			return false;
		queryWasForced = false;

		// If the user chose another autoscan, then do so
		if (urlStubPrefix == "scanagain")
		{
			if (AutomaticallyFindURL(urlStubPrefix, urlStubSuffix))
			{
				url = MakeFullURL(urlStubPrefix, urlStubSuffix);
				return true;
			}
		}
		else
		{
			// Check to see if their URL is OK
			if (TrySingleURL(urlStubPrefix, urlStubSuffix))
			{
				url = MakeFullURL(urlStubPrefix, urlStubSuffix);
				return true;
			}
		}

		// Nothing worked, so loop round to ask the user again
	}
}
