// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - May 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
#include "priv.h"
#include "ShellExt.h"
#include <string>
#include "../CVSGlue/PreserveChdir.h"
#include "../Utils/PathUtils.h"
#include "../Utils/UnicodeStrings.h"
#include "../Utils/TortoiseRegistry.h"

#include "../ContextMenus/HasMenu.h"
#include "../ContextMenus/MenuDescription.h"
#include "../ContextMenus/DefaultMenus.h"

#include "TortoiseShellRes.h"

void CShellExt::ContextMenuInitialise()
{
	FillInDefaultMenus(myMenus);

	// Index by verb
	for (unsigned int i = 0; i < myMenus.size(); ++i)
	{
		std::string verb = myMenus[i].GetVerb();
		myVerbMap[verb] = i;
	}
}

void CShellExt::ContextMenuFinalise()
{

}
STDMETHODIMP CShellExt::QueryContextMenu(HMENU hMenu,
                                         UINT indexMenu,
                                         UINT idCmdFirst,
                                         UINT idCmdLast,
                                         UINT uFlags)
{
    PreserveChdir preserveChdir;

	if (
		AreShellExtTargetsAvail() &&
		(((uFlags & 0x000F) == CMF_NORMAL) || (uFlags & CMF_EXPLORE))
	)
    {
	    UINT idCmd = idCmdFirst;

		// Separator before
        InsertMenu(hMenu, indexMenu++, MF_SEPARATOR|MF_BYPOSITION, 0, NULL);

		// Delete any existing "C&VS" submenu.  This is a workaround for an 
		// Explorer bug - it removes all normal menu items, but not submenu
		// items that we added previously.  This is on the main File menu -
		// for popup context menus, the menu must be generated afresh, so
		// everything is fine.
		int count = GetMenuItemCount(hMenu);
		for (int j = 0; j < count; ++j)
		{
			char buf[4096];
			GetMenuString(hMenu, j, buf, 4095, MF_BYPOSITION);
			if (std::string(buf) == "C&VS")
				DeleteMenu(hMenu, j, MF_BYPOSITION);
		}

		// This is deleted because DestroyMenu is recursive - hopefully
		// Explorer just calls that!
		HMENU subMenu = CreateMenu();
		int indexSubMenu = 0;

		bool hasSubMenu = false;
		// TODO: Better separator mechanism
		bool previousSubMenuWasSeparator = true; // stop two adjacent separators
		for (unsigned int i = 0; i < myMenus.size(); ++i)
		{
			MenuDescription& menuDescription = myMenus[i];
			if (HasMenu(menuDescription.GetFlags(), m_Files))
			{
				// Add item to either sub menu or main menu
				if (menuDescription.GetFlags() & ON_SUB_MENU)
				{
					if (menuDescription.GetVerb().empty())
					{
						if (!previousSubMenuWasSeparator)
						{
							InsertMenu(subMenu, indexSubMenu++, MF_SEPARATOR|MF_BYPOSITION, 0, NULL);
							idCmd++;
							previousSubMenuWasSeparator = true;
						}
					}
					else
					{
						std::string menuText = menuDescription.GetMenuText();
						InsertMenu(subMenu, indexSubMenu++, MF_STRING|MF_BYPOSITION, idCmd++, menuText.c_str());
						hasSubMenu = true;
						previousSubMenuWasSeparator = false;
						// We store the relative and absolute diameter
						// (drawitem callback uses absolute, others relative)
						myIDMap[idCmd - idCmdFirst - 1] = i;
						myIDMap[idCmd - 1] = i;
					}
				}
				else
				{
					if (menuDescription.GetVerb().empty())
					{
						InsertMenu(hMenu, indexMenu++, MF_SEPARATOR|MF_BYPOSITION, 0, NULL);
						idCmd++;
					}
					else
					{
						int flags = MF_STRING|MF_BYPOSITION;
						if (TortoiseRegistry::ReadInteger("ContextIcons", 1) > 0)
							flags |= MF_OWNERDRAW;

						std::string menuText = "CVS " + menuDescription.GetMenuText();
						InsertMenu(hMenu, indexMenu++, flags, idCmd++, menuText.c_str());
						// We store the relative and absolute diameter
						// (drawitem callback uses absolute, others relative)
						myIDMap[idCmd - idCmdFirst - 1] = i;
						myIDMap[idCmd - 1] = i;
					}
				}
			}
		}

		// Add sub menu to main menu
		if (hasSubMenu)
			InsertMenu(hMenu, indexMenu++, MF_POPUP|MF_BYPOSITION|MF_STRING, (UINT)subMenu, "C&VS");
		else
			DestroyMenu(subMenu);

		// Separator after
        InsertMenu(hMenu, indexMenu++, MF_SEPARATOR|MF_BYPOSITION, 0, NULL);
        
		// Return number of menu items that we added
        return ResultFromShort(idCmd - idCmdFirst);
   }

   return NOERROR;
}


// This is called when you invoke a command on the menu:
STDMETHODIMP CShellExt::InvokeCommand(LPCMINVOKECOMMANDINFO lpcmi)
{
	PreserveChdir preserveChdir;

	HRESULT hr = E_INVALIDARG;

    // If HIWORD(lpcmi->lpVerb) then we have been called programmatically
    // and lpVerb is a command that should be invoked.  Otherwise, the shell
    // has called us, and LOWORD(lpcmi->lpVerb) is the menu ID the user has
    // selected.  Actually, it's (menu ID - idCmdFirst) from QueryContextMenu().
	MenuDescription* menuDescription = NULL;
	if (!HIWORD(lpcmi->lpVerb))
    {
		if (!AreShellExtTargetsAvail())
		{
		    MessageBox(lpcmi->hwnd, "Command invoked without IShellExtInit::Initialize", "TortoiseCVS", MB_OK);
			return hr;
		}

        UINT idCmd = LOWORD(lpcmi->lpVerb);

		// See if we have a handler interface for this id
		if (myIDMap.find(idCmd) != myIDMap.end())
			menuDescription = &myMenus[myIDMap[idCmd]];
    }
	else
	{
		// See if we have a handler interface for this verb
		if (myVerbMap.find(lpcmi->lpVerb) != myVerbMap.end())
			menuDescription = &myMenus[myVerbMap[lpcmi->lpVerb]];
	}

	// Now we have our handler, we can perform the action
	if (menuDescription)
	{
		hr = NOERROR;
		menuDescription->Perform(m_Files, lpcmi->hwnd);
	}

    return hr;
}



// This is for the status bar and things like that:
STDMETHODIMP CShellExt::GetCommandString(UINT idCmd,
	                                         UINT uFlags,
                                         UINT FAR *reserved,
                                         LPSTR pszName,
                                         UINT cchMax)
{   
	// See if we have a handler interface for this id
	if (myIDMap.find(idCmd) == myIDMap.end())
	{
		return E_INVALIDARG;
	}

	MenuDescription& iface = myMenus[myIDMap[idCmd]];
	HRESULT hr = E_INVALIDARG;

	// Return the appropriate help text or verb for the command handler
    switch(uFlags)
    {
        case GCS_HELPTEXTA:
			{
				std::string help = iface.GetHelpText();
				lstrcpynA(pszName, help.c_str(), cchMax);
				hr = S_OK;
				break; 
			}

        case GCS_HELPTEXTW: 
			{
				wide_string help = MultibyteToWide(iface.GetHelpText());
				lstrcpynW((LPWSTR)pszName, help.c_str(), cchMax); 
				hr = S_OK;
				break; 
			}

        case GCS_VERBA:
			{
				std::string verb = iface.GetVerb();
				lstrcpynA(pszName, verb.c_str(), cchMax); 
				hr = S_OK;
				break; 
			}

        case GCS_VERBW:
			{
				wide_string verb = MultibyteToWide(iface.GetVerb());
				lstrcpynW((LPWSTR)pszName, verb.c_str(), cchMax); 
				hr = S_OK;
				break; 
			}
   
        default:
            hr = S_OK;
            break; 
    }
    return hr;
}

// For owner draw (icons, and maybe one day more!)

STDMETHODIMP CShellExt::HandleMenuMsg(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	LRESULT res;
	return HandleMenuMsg2(uMsg, wParam, lParam, &res);
}

STDMETHODIMP CShellExt::HandleMenuMsg2(UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT *pResult)
{
	// Sometimes, pResult is NULL...
	LRESULT res;
	if (pResult == NULL)
		pResult = &res;
	
	// See what message we have
	if (uMsg == WM_MEASUREITEM)
	{
		MEASUREITEMSTRUCT* lpmis = (MEASUREITEMSTRUCT*)lParam;
		lpmis->itemWidth = 150; 
		lpmis->itemHeight = 20; 
	}
	else if (uMsg == WM_DRAWITEM)
	{
		// Choose icon type
		int iconType =  TortoiseRegistry::ReadInteger("ContextIcons", 1);
		bool smallIcon = true;
		if (iconType == 2)
			smallIcon = true;
		else if (iconType == 3)
			smallIcon = false;
		else
		{
			// Decide whether to use small or large menu icon based on
			// what I think I know about menu sizes on the OSes.
			//		Windows NT/2000 - small icon
			//		Windows 95/98 - large icon
			// For now, see if this works best for people!
			OSVERSIONINFO osv;
			osv.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
			if( !GetVersionEx(&osv) )
				MessageBox(NULL, "GetVersionEx failed", "TortoiseCVS", MB_OK);
			if (osv.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
				smallIcon = false;
		}
			
		DRAWITEMSTRUCT* lpdis = (DRAWITEMSTRUCT*)lParam;
		
		// Fine positioning
		const int DEFBITMAPSIZEX = smallIcon ? 13 : 20;
		const int DEFBITMAPSIZEY = smallIcon ? 13 : 16;
		const int BITMAPOFFSET = 0;
		const int TEXTOFFSET = smallIcon ? 5 : 0; // more means further left
		
		// if not called by a menu, exit
		if (lpdis->CtlType != ODT_MENU)
			return S_OK;
			
		// retrieve the data associated with the menu item
		MenuDescription& iface = myMenus[myIDMap[lpdis->itemID]];
		std::string itemText = "CVS " + iface.GetMenuText();
		char *szItem = (char*)itemText.c_str();
		
		// manage how to draw
		if (lpdis->itemAction & (ODA_DRAWENTIRE|ODA_SELECT))
		{
			int ix, iy; 
			RECT rt, rtTemp;
			SIZE size;
			
			// select the appropriate text colors
			if (lpdis->itemState & ODS_SELECTED)
			{
				COLORREF crText;
				if (lpdis->itemState & ODS_GRAYED)
					crText = SetTextColor( lpdis->hDC, RGB(128,128,128) );
				else
					crText = SetTextColor( lpdis->hDC, GetSysColor(COLOR_HIGHLIGHTTEXT) );
				COLORREF crBack = SetBkColor( lpdis->hDC, GetSysColor(COLOR_HIGHLIGHT) );
			}
			
			
			// get bitmap bits and paint
			CopyRect(&rtTemp, &(lpdis->rcItem));
			
			// Load icon in
			LPCTSTR resource;
			if (smallIcon)
				resource = MAKEINTRESOURCE(IDI_TORTOISE_13);
			else
				resource = MAKEINTRESOURCE(IDI_TORTOISE_20);
			HICON hIcon = (HICON) LoadImage(GetModuleHandle("TortoiseShell"), resource, IMAGE_ICON, 0, 0, LR_DEFAULTCOLOR);
			
			// Convert HICON to HBITMAP
			ICONINFO ii;
			GetIconInfo(hIcon, &ii);
			HBITMAP hbmItem = ii.hbmColor; 
			if (hbmItem)
			{
				BITMAP bm;
				GetObject(hbmItem, sizeof(bm), &bm);
				
				ix = lpdis->rcItem.left+BITMAPOFFSET;
				int tempY = lpdis->rcItem.top + ((lpdis->rcItem.bottom-lpdis->rcItem.top)-bm.bmHeight)/2;
				SetRect(&rt, ix, tempY, ix+DEFBITMAPSIZEX, tempY+DEFBITMAPSIZEY);
				
				ExtTextOut(lpdis->hDC, 0, 0, ETO_CLIPPED|ETO_OPAQUE, &rtTemp, NULL, 0, (LPINT)NULL);
				rtTemp.left = rt.right;
				
				HDC hDCTemp = CreateCompatibleDC(lpdis->hDC);
				SelectObject(hDCTemp, hbmItem);
				DrawIconEx(lpdis->hDC, rt.left, rt.top, hIcon, bm.bmWidth, bm.bmHeight,
					0, NULL, DI_NORMAL);

				DeleteDC(hDCTemp);
				DeleteObject(hbmItem);
			}
			ix = rt.right + GetSystemMetrics(SM_CXFRAME) - TEXTOFFSET;
			
			// draw text (one line centered vertically)
			GetTextExtentPoint32(lpdis->hDC, szItem, lstrlen(szItem), &size);
			iy = ((lpdis->rcItem.bottom-lpdis->rcItem.top)-size.cy)/2;
			iy = lpdis->rcItem.top + (iy>=0 ?iy :0);
			SetRect(&rt, ix+1, iy, lpdis->rcItem.right - TEXTOFFSET, lpdis->rcItem.bottom);
			ExtTextOut(lpdis->hDC, ix, iy, ETO_CLIPPED|ETO_OPAQUE, &rtTemp, NULL, 0, (LPINT)NULL);
			if (lpdis->itemState & ODS_GRAYED)
			{			
				SetBkMode(lpdis->hDC, TRANSPARENT);
				OffsetRect( &rt, 1, 1 );
				SetTextColor( lpdis->hDC, RGB(255,255,255) );
				DrawText( lpdis->hDC, szItem, lstrlen(szItem), &rt, DT_LEFT|DT_EXPANDTABS );
				OffsetRect( &rt, -1, -1 );
				SetTextColor( lpdis->hDC, RGB(128,128,128) );
				DrawText( lpdis->hDC, szItem, lstrlen(szItem), &rt, DT_LEFT|DT_EXPANDTABS );			 
			}
			else
				DrawText( lpdis->hDC, szItem, lstrlen(szItem), &rt, DT_LEFT|DT_EXPANDTABS );
		}
	}
	
	return S_OK;
}
