// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - May 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// Implementation of IShellIconOverlayIdentifier interface

#include "priv.h"
#include "ShellExt.h"
#include "../CVSGlue/CVSStatus.h"
#include "../Utils/UnicodeStrings.h"
#include "../CVSGlue/PreserveChdir.h"
#include "../Utils/PathUtils.h"
#include "../Utils/TortoiseUtils.h"
#include "../Utils/TortoiseDebug.h"
#include "../Utils/TortoiseRegistry.h"

STDMETHODIMP CShellExt::GetOverlayInfo(LPWSTR pwszIconFile, int cchMax, int *pIndex, DWORD *pdwFlags)
{
	OSVERSIONINFO osv;
	osv.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if( !GetVersionEx(&osv) )
	{
	    MessageBox(NULL, "GetVersionEx failed", "TortoiseCVS", MB_OK);
		return S_FALSE;
	}
	// Under NT (and 95?), file open dialog crashes apps upon shutdown,
	// so we disable our icon overlays unless we are in Explorer process
	// space.
	bool allowExplorer = osv.dwMajorVersion > 4 || // allow anything major > 4
			(osv.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS &&
			 osv.dwMajorVersion == 4 && osv.dwMinorVersion > 0); // plus Windows 98/Me
	// Test if we are in Explorer
	if (!allowExplorer)
	{
		char buf[_MAX_PATH + 1];
		GetModuleFileName(NULL, buf, _MAX_PATH);
		std::string filename = buf;
		// Does this need case insensitivity?
		if (filename.find("explorer", 0) == std::string::npos)
			return S_FALSE;
	}

	// Get folder icons are in (same as DLL)
	std::string folder = GetTortoiseFolder();
	wide_string folderWide = MultibyteToWide(folder);
	wcsncpy(pwszIconFile, folderWide.c_str(), cchMax);
	cchMax -= folderWide.size() + 1;

	// Add name of appropriate icon
	if (m_TortoiseClass == TORTOISE_OLE_CONFLICT)
		wcsncat(pwszIconFile, L"TortoiseConflict.ico", cchMax);
	else if (m_TortoiseClass == TORTOISE_OLE_CHANGED)
		wcsncat(pwszIconFile, L"TortoiseChanged.ico", cchMax);
	else if (m_TortoiseClass == TORTOISE_OLE_NOTINCVS)
		wcsncat(pwszIconFile, L"TortoiseNotInCVS.ico", cchMax);
	else if (m_TortoiseClass == TORTOISE_OLE_INCVS)
		wcsncat(pwszIconFile, L"TortoiseInCVS.ico", cchMax);
	else
		return S_FALSE;

	*pIndex = 0;
	*pdwFlags = ISIOI_ICONFILE;
	return S_OK;
};

STDMETHODIMP CShellExt::GetPriority(int *pPriority)
{
	if (m_TortoiseClass == TORTOISE_OLE_CONFLICT)
		*pPriority = 60;
	else if (m_TortoiseClass == TORTOISE_OLE_CHANGED)
		*pPriority = 30;
	else if (m_TortoiseClass == TORTOISE_OLE_NOTINCVS)
		*pPriority = 90;
	else if (m_TortoiseClass == TORTOISE_OLE_INCVS)
		*pPriority = 10;
	else 
		return S_FALSE;
	return S_OK;
}

STDMETHODIMP CShellExt::IsMemberOf(LPCWSTR pwszPath, DWORD /* dwAttrib */)
{
	PreserveChdir preserveChdir;
	std::string path = WideToMultibyte(wide_string(pwszPath));

	CVSStatus::FileStatus status;
	// The recursive version is too slow for use, although it
	// is beautiful.  It lets you see changed directories - i.e.
	// ones which have a file within them that has changed (or
	// been added/removed) recursively.
	if (TortoiseRegistry::ReadBoolean("Show Changed Directories", false))
		status = CVSStatus::GetFileStatusRecursive(path);
	else
		status = CVSStatus::GetFileStatus(path);

	if (status == CVSStatus::STATUS_NOWHERENEAR_CVS)
		return S_FALSE;

	if (m_TortoiseClass == TORTOISE_OLE_CONFLICT && status == CVSStatus::STATUS_CONFLICT)
		return S_OK;
	else if (m_TortoiseClass == TORTOISE_OLE_CHANGED && status == CVSStatus::STATUS_CHANGED)
		return S_OK;
	else if (m_TortoiseClass == TORTOISE_OLE_NOTINCVS && status == CVSStatus::STATUS_NOTINCVS)
		return S_OK;
	else if (m_TortoiseClass == TORTOISE_OLE_INCVS && (status == CVSStatus::STATUS_INCVS || status == CVSStatus::STATUS_OUTERDIRECTORY))
		return S_OK;

	return S_FALSE;
}

