// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - May 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
#include "priv.h"
#include "ShellExt.h"
#include <string>
#include <sstream>
#include "../CVSGlue/CVSStatus.h"
#include "../CVSGlue/PreserveChdir.h"
#include "../Utils/PathUtils.h"
#include "../Utils/UnicodeStrings.h"
#include "../Utils/TimeUtils.h"
#include "../Utils/TortoiseUtils.h"
#include "../Utils/PropSheetMsg.h"
#include "../Utils/TortoiseRegistry.h"
#include "../ContextMenus/HasMenu.h"
#include "TortoiseShellRes.h"
#include "PropSheet.h"

/////////////////////////////////////////////////////////////////////////////
// List box defines

enum
{
	COL_REVISION,
	COL_CHECKEDIN,
	COL_AUTHOR,
	COL_TYPE,
	COL_BRANCH,
	NUM_COLUMNS
};

static char * pszLabels[] = {
	"Revision",
	"Checked in",
	"Author",
	"Type",
	"Branch"
};

static char * pszKeywordStyles[] = {
	"Normal",
	"Only keyword names",
	"Include locker",
	"From last checkin",
	"Only values"
};
#define NUM_KEYWORD_STYLES (sizeof(pszKeywordStyles)/sizeof(pszKeywordStyles[0]))

enum {
	keyNormal,
	keyOnlyNames,
	keyShowLocker,
	keyShowPrevious,
	keyOnlyValues,
	keyBinary
};

static char * pszCVSStatus[] = {
	"Not in CVS",	// STATUS_NOWHERENEAR_CVS
	"Unmodified",	// STATUS_INCVS
	"Modified",		// STATUS_CHANGED
	"Needs merge",	// STATUS_CONFLICT
	"Not in CVS",	// STATUS_NOTINCVS
	"Module"		// STATUS_OUTERDIRECTORY
};

/////////////////////////////////////////////////////////////////////////////
// Nonmember function prototypes

BOOL CALLBACK PageProc (HWND, UINT, WPARAM, LPARAM);
UINT CALLBACK PropPageCallbackProc ( HWND hwnd, UINT uMsg, LPPROPSHEETPAGE ppsp );
int CALLBACK CompareFunc (LPARAM, LPARAM, LPARAM);

BOOL IsAnyItemSelected (HWND);

/////////////////////////////////////////////////////////////////////////////
// CShellExt member functions

STDMETHODIMP CShellExt::AddPages (LPFNADDPROPSHEETPAGE lpfnAddPage,
    LPARAM lParam)
{
	// Only show property sheet if there's something worth showing
	if (!HasMenu(
			EXACTLY_ONE|TYPE_INCVS|TYPE_CHANGED|
				TYPE_CONFLICT|TYPE_OUTERDIRECTORY,
			m_Files
		)
	)
	{
	    return NOERROR;
	}

    PROPSHEETPAGE psp;
    HPROPSHEETPAGE hPage;
	CVSRevisionsPropSheet *sheet = new CVSRevisionsPropSheet(m_Files[0]);

    psp.dwSize = sizeof (psp);
    psp.dwFlags = PSP_USEREFPARENT | PSP_USETITLE | PSP_USEICONID | PSP_USECALLBACK;
    psp.hInstance = g_hmodThisDll;
    psp.pszTemplate = "Revisions";
	psp.pszIcon = MAKEINTRESOURCE(IDI_PROPTAB);
    psp.pszTitle = "CVS";
    psp.pfnDlgProc = (DLGPROC) PageProc;
    psp.lParam = (LPARAM) sheet;
	psp.pfnCallback = PropPageCallbackProc;
    psp.pcRefParent = &g_cRefThisDll;

    hPage = CreatePropertySheetPage (&psp);

    if (hPage != NULL)
        if (!lpfnAddPage (hPage, lParam))
		{
			delete sheet;
            DestroyPropertySheetPage (hPage);
		}

	return NOERROR;
}

STDMETHODIMP CShellExt::ReplacePage (UINT uPageID,
    LPFNADDPROPSHEETPAGE lpfnReplaceWith, LPARAM lParam)
{
    return E_FAIL;
}

/////////////////////////////////////////////////////////////////////////////
// Dialog procedures and other callback functions

BOOL CALLBACK PageProc (HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
	CVSRevisionsPropSheet * sheet;

	if (uMessage == WM_INITDIALOG)
	{
		sheet = (CVSRevisionsPropSheet*) ((LPPROPSHEETPAGE) lParam)->lParam;
        SetWindowLong (hwnd, GWL_USERDATA, (LONG) sheet);
		sheet->SetHwnd(hwnd);
	}
	else
	{
		sheet = (CVSRevisionsPropSheet*) GetWindowLong (hwnd, GWL_USERDATA);
	}

	if (sheet != 0L)
		return sheet->PageProc(uMessage, wParam, lParam);
	else
	    return FALSE;
}

UINT CALLBACK PropPageCallbackProc ( HWND hwnd, UINT uMsg, LPPROPSHEETPAGE ppsp )
{
	//
	// Delete the filename before closing.
	//
	if ( PSPCB_RELEASE == uMsg )
	{
		CVSRevisionsPropSheet* sheet = (CVSRevisionsPropSheet*) ppsp->lParam;
        if (sheet != NULL)
            delete sheet;
	}
	return 1;
}

int CALLBACK CompareFunc (LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	CVSRevisionsPropSheet * sheet = (CVSRevisionsPropSheet *) lParamSort;
	return sheet->Compare(lParam1, lParam2);
}

// *********************** CVSRevisionsPropSheet *************************

CVSRevisionsPropSheet::CVSRevisionsPropSheet(const std::string &newFilename)
	: filename(newFilename)
{
}
    
CVSRevisionsPropSheet::~CVSRevisionsPropSheet()
{
}

void CVSRevisionsPropSheet::SetHwnd(HWND newHwnd)
{
	hwnd = newHwnd;
	hwndListView = GetDlgItem (hwnd, IDD_LISTVIEW);
}

BOOL CVSRevisionsPropSheet::PageProc (UINT uMessage, WPARAM wParam, LPARAM lParam)
{
    NM_LISTVIEW* pnmlv;
	NMLVDISPINFO* plvdi;
    UINT nID;

    switch (uMessage) {
    
    case WM_INITDIALOG:
        //
        // Initialize the ListView control.
        //
		InitWorkfileView();
        InitListView ();
        return TRUE;    

    case WM_COMMAND:
        //
        // Process button clicks.
        //
        nID = LOWORD (wParam);

#if 0
		// TODO[JR]: add context menu to ListView
        switch (nID) {
        }
#endif
        break;

	case WM_COPYDATA:
		{
			PropSheetMsgChannel sender((PCOPYDATASTRUCT) lParam);
			PropSheetMsgHeader header;
			if (!sender.Get(header))
				/* ignore a bad message */;
			else if (PropSheetSymbolicMsg::IsMyHeader(header))
			{
				PropSheetSymbolicMsg symbolic;
				sender.Get(symbolic);
				AddRevision(CVSRevisionData(
					symbolic.isBranch? CVSRevisionData::Branch : CVSRevisionData::Symbolic,
					symbolic.revisionNumber, symbolic.tag, "",
					0, ""
					)
				);
			}
			else if (PropSheetRevisionMsg::IsMyHeader(header))
			{
				PropSheetRevisionMsg revision;
				sender.Get(revision);
				AddRevision(CVSRevisionData(CVSRevisionData::Revision,
					revision.revisionNumber, "", "",
					&revision.timestamp, revision.author));
			}
		}
		break;

    case WM_NOTIFY:
        //
        // Respond to notifications.
        //    
        switch (((NMHDR*) lParam)->code) {

        case PSN_APPLY:
        case PSN_RESET:
            SaveRegistryData();
            break;

        case LVN_COLUMNCLICK:
            pnmlv = (NM_LISTVIEW*) lParam;
			sort.DoClick(pnmlv->iSubItem);
            ListView_SortItems (hwndListView, CompareFunc, (LPARAM) this);
            break;

        case LVN_GETDISPINFO:
			plvdi = (NMLVDISPINFO*) lParam;
            revisions[plvdi->item.lParam].GetDispInfo (plvdi);
            break;

        case LVN_ITEMCHANGED:
            pnmlv = (NM_LISTVIEW*) lParam;
//TODO[JR]            EnableWindow (GetDlgItem (hwnd, IDD_EXTRACT),
//                IsAnyItemSelected (hwndListView));
            break;
         }
        SetWindowLong (hwnd, DWL_MSGRESULT, FALSE);
        return TRUE;        

    case WM_DESTROY:
        return TRUE;
    }
    return FALSE;
}


void CVSRevisionsPropSheet::UpdateControls()
{
	// display keyword
	HWND hwndCombo = GetDlgItem(hwnd, IDC_KEY_STYLE);
	int box = keyNormal;

	if (CVSStatus::HasOptions(filename)) {
		std::string option = CVSStatus::GetOptions(filename);

		if (option == "-kb")		box = keyBinary;
		else if (option == "-ko")	box = keyShowPrevious;
		else if (option == "-kkvl")	box = keyShowLocker;
		else if (option == "-kk")	box = keyOnlyNames;
		else if (option == "-kv")	box = keyOnlyValues;
	}

	// set binary check box
	CheckDlgButton(hwnd, IDC_KEY_BINARY, box==keyBinary);
	EnableWindow(GetDlgItem(hwnd, IDC_KEY_BINARY), FALSE);

	// set keyword style combo box
	EnableWindow(hwndCombo, box != keyBinary);
	SendMessage(hwndCombo, CB_RESETCONTENT, 0, 0);
	if (box != keyBinary) {
		// fill combo box content
		for (int i=0; i<NUM_KEYWORD_STYLES; i++) 
		{
			SendMessage(hwndCombo, CB_ADDSTRING, 
				0, (LPARAM) pszKeywordStyles[i]);
		}
		// set keyword combo box value
		SendMessage(hwndCombo, CB_SETCURSEL, (WPARAM) box, 0);
	}
	EnableWindow(hwndCombo, FALSE);
}

void CVSRevisionsPropSheet::InitWorkfileView()
{
	// show revision number
	if (CVSStatus::HasRevisionNumber(filename)) {
		SetDlgItemText (hwnd, IDC_REVISION, CVSStatus::GetRevisionNumber(filename).c_str());
	}

	// show sticky tag or date
	if (CVSStatus::HasStickyTag(filename)) {
		SetDlgItemText (hwnd, IDC_TAG, CVSStatus::GetStickyTag(filename).c_str());
	}
	else if (CVSStatus::HasStickyDate(filename)) {
		SetDlgItemText (hwnd, IDC_TAG, CVSStatus::GetStickyDate(filename).c_str());
	}
	else {
		SetDlgItemText (hwnd, IDC_TAG, "Head");
	}

	// show timestamp
	if (CVSStatus::HasTimestamp(filename)) {
		std::string ascTimestamp = CVSStatus::GetTimestamp(filename);
		char szDate[100];

		if (asctime_to_local_DateTimeFormatted(ascTimestamp.c_str(), szDate, sizeof(szDate)))
		{
			SetDlgItemText (hwnd, IDC_TIMESTAMP, szDate);
		}
	}

	// show file status
	CVSStatus::FileStatus status = CVSStatus::GetFileStatus(filename);
	SetDlgItemText(hwnd, IDC_STATUS, pszCVSStatus[status]);

	// show keyword expansion
	UpdateControls();
}

void CVSRevisionsPropSheet::InitListView()
{
    // Read the column widths from the registry and create the columns.
    //
	GetRegistryData();

	/* get the first revision type icon */
	HICON hicon;
	hicon = (HICON) LoadImage(GetModuleHandle("TortoiseShell"), MAKEINTRESOURCE(IDI_TYPEREV), IMAGE_ICON, 0, 0, LR_DEFAULTCOLOR);

	/* get the size of this icon */
	BITMAP bm;
	ICONINFO iconInfo;
	GetIconInfo(hicon, &iconInfo);
	GetObject(iconInfo.hbmColor, sizeof(bm), &bm);
	if (iconInfo.hbmColor != 0)
		DeleteObject(iconInfo.hbmColor);
	if (iconInfo.hbmMask != 0)
		DeleteObject(iconInfo.hbmMask);

	/* build an image list from icon resources */
	hwndImageList = ImageList_Create(bm.bmWidth, bm.bmHeight, ILC_COLOR|ILC_MASK, 0, 3);
	ImageList_AddIcon(hwndImageList, hicon);
	hicon = (HICON) LoadImage(GetModuleHandle("TortoiseShell"), MAKEINTRESOURCE(IDI_TYPESYM), IMAGE_ICON, 0, 0, LR_DEFAULTCOLOR);
	ImageList_AddIcon(hwndImageList, hicon);
	hicon = (HICON) LoadImage(GetModuleHandle("TortoiseShell"), MAKEINTRESOURCE(IDI_TYPEBRA), IMAGE_ICON, 0, 0, LR_DEFAULTCOLOR);
	ImageList_AddIcon(hwndImageList, hicon);
	ListView_SetImageList(hwndListView, hwndImageList, LVSIL_SMALL);
 
	/* setup the columns */
    LVCOLUMN lvc;
    for (unsigned int i=0; i<NUM_COLUMNS; i++)
	{
		lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT;
		lvc.fmt = LVCFMT_LEFT;
		lvc.pszText = pszLabels[i];
		lvc.cx = nWidths[i];

		if (i == COL_REVISION)
			lvc.fmt |= LVCFMT_IMAGE;
        ListView_InsertColumn (hwndListView, i, &lvc);
    }

	// start filling in the list box
	LaunchTortoiseAct("CVSLogPropSheet", filename, hwnd);
}

void CVSRevisionsPropSheet::AddRevision(CVSRevisionData const & newRevision)
{
	revisions.push_back(newRevision);

	/* setup LVITEM */
    LVITEM lvi;
    lvi.mask = LVIF_TEXT | LVIF_PARAM | LVIF_IMAGE;
    lvi.iItem = revisions.size() - 1;
    lvi.iSubItem = 0;
    lvi.pszText = LPSTR_TEXTCALLBACK;
	lvi.lParam = lvi.iItem;
	lvi.iImage = I_IMAGECALLBACK;

	/* find sorted position for LVITEM */
	int numItems = ListView_GetItemCount(hwndListView);
	int i;
	for (i=0; i<numItems; i++)
	{
		LVITEM cur;
		cur.mask = LVIF_PARAM;
		cur.iItem = i;
		cur.iSubItem = 0;
		ListView_GetItem(hwndListView, &cur);
		if (Compare(cur.lParam, lvi.lParam) > 0)
			break;
	}

	/* insert LVITEM */
	lvi.iItem = i;
    ListView_InsertItem (hwndListView, &lvi);
}

void CVSRevisionsPropSheet::SaveRegistryData()
{
    for (int i=0; i<NUM_COLUMNS; i++) {
		std::stringstream name;

		name << "Col" << i << "Width";
		TortoiseRegistry::WriteInteger(name.str(), ListView_GetColumnWidth (hwndListView, i));
    }

	/* save sort criteria */
	sort.SaveRegistryData();
}

void CVSRevisionsPropSheet::GetRegistryData()
{
	nWidths.resize(NUM_COLUMNS);

	/* read saved values from registry */
    for (int i=0; i<NUM_COLUMNS; i++) {
		std::stringstream name;

		name << "Col" << i << "Width";
		nWidths[i] = TortoiseRegistry::ReadInteger(name.str(), i==0? 128 : 64);
    }

	/* get sort criteria */
	sort.GetRegistryData();
}

int CVSRevisionsPropSheet::Compare(LPARAM lParam1, LPARAM lParam2) const
{
	const CVSRevisionData * rev1 = &revisions[lParam1];
	const CVSRevisionData * rev2 = &revisions[lParam2];
	return sort.Compare(*rev1, *rev2);
}


// *********************** CVSRevisionDataSortCriteria *************************
void CVSRevisionDataSortCriteria::DoClick(int column)
{
	if (column == COL_TYPE)
		groupSameTypes = !groupSameTypes;
	else if (column == key)
		ascending = !ascending;
	else
	{
		key = column;
		ascending = true;
	}
}

void CVSRevisionDataSortCriteria::GetRegistryData()
{
	key = TortoiseRegistry::ReadInteger("PropSortKey", 0);
	ascending = TortoiseRegistry::ReadBoolean("PropSortAscending", true);
	groupSameTypes = TortoiseRegistry::ReadBoolean("PropSortGroupSameTypes", true);
}

void CVSRevisionDataSortCriteria::SaveRegistryData() const
{
	TortoiseRegistry::WriteInteger("PropSortKey", key);
	TortoiseRegistry::WriteBoolean("PropSortAscending", ascending);
	TortoiseRegistry::WriteBoolean("PropeSortGroupSameTypes", groupSameTypes);
}

int CVSRevisionDataSortCriteria::Compare(const CVSRevisionData::Type t1, const CVSRevisionData::Type t2) const
{
	int ret;

	if (t1 > t2)
		ret = 1;
	else if (t1 < t2)
		ret = -1;
	else
		ret = 0;

	return ret;
}

int CVSRevisionDataSortCriteria::Compare(const std::string &s1, const std::string &s2) const
{
	if (ascending)
		return s1.compare(s2);
	else
		return -s1.compare(s2);
}

int CVSRevisionDataSortCriteria::Compare(const LONGLONG &t1, const LONGLONG &t2) const
{
	int ret;

	if (t1 > t2)
		ret = 1;
	else if (t1 < t2)
		ret = -1;
	else
		ret = 0;

	if (ascending)
		return ret;
	else
		return -ret;
}

int CVSRevisionDataSortCriteria::Compare(const CVSRevisionData &rev1, const CVSRevisionData &rev2) const
{
	int res = 0;
	int compareTypes = Compare(rev1.type, rev2.type);

	if (groupSameTypes)
		res = compareTypes;
	if (res != 0)
		return res;

    switch (key) {
    case COL_REVISION:
		if (groupSameTypes)
		{
			if (rev1.type == CVSRevisionData::Revision)
				res = Compare(rev1.revision, rev2.revision);
			else
				res = Compare(rev1.tag, rev2.tag);
		}
		else
		{
			res = Compare(rev1.revision, rev2.revision);
			if (res == 0)
				res = compareTypes;
			if (res == 0)
				res = Compare(rev1.tag, rev2.tag);
		}
		break;
	case COL_BRANCH:
		res = Compare(rev1.branch, rev2.branch);
		break;
	case COL_CHECKEDIN:
		res = Compare(rev1.checkedInOrdinal, rev2.checkedInOrdinal);
		break;
	case COL_AUTHOR:
		res = Compare(rev1.author, rev2.author);
		break;
	}
	if (res != 0)
		return res;

	if (!groupSameTypes)
		res = compareTypes;
	return res;
}


// *********************** CVSRevisionData *************************
CVSRevisionData::CVSRevisionData(CVSRevisionData::Type _type, const std::string & _revision, const std::string & _tag,
		const std::string & _branch, const struct tm * _checkedIn, const std::string & _author) :
	type(_type), revision(_revision), tag(_tag), branch(_branch), author(_author)
{
	char szDate[100];

	if (_checkedIn != 0 && tm_to_local_DateTimeFormatted(_checkedIn, szDate, sizeof(szDate), false))
	{
		SYSTEMTIME utcTime;
		FILETIME fileTime;
		checkedIn = szDate;
		tm_to_SYSTEMTIME(_checkedIn, &utcTime);
		SystemTimeToFileTime(&utcTime, &fileTime);
		checkedInOrdinal = * (LONGLONG *) &fileTime;
	}
	else
	{
		checkedIn = "";
		checkedInOrdinal = 0;
	}
}

void CVSRevisionData::GetDispInfo (NMLVDISPINFO* plvdi)
{
    if (plvdi->item.mask & LVIF_TEXT) {
        switch (plvdi->item.iSubItem) {
        case COL_REVISION:
			switch (type)
			{
			case Revision:
				lstrcpy (plvdi->item.pszText, revision.c_str());
				break;
			case Symbolic:
			case Branch:
				lstrcpy (plvdi->item.pszText, tag.c_str());
				break;
			}
            break;

        case COL_BRANCH:
            lstrcpy (plvdi->item.pszText, branch.c_str());
            break;

        case COL_CHECKEDIN:
            lstrcpy (plvdi->item.pszText, checkedIn.c_str());
            break;

        case COL_AUTHOR:
            lstrcpy (plvdi->item.pszText, author.c_str());
            break;

		case COL_TYPE:
			switch (type)
			{
			case Revision:
				lstrcpy (plvdi->item.pszText, "Revision");
				break;
			case Symbolic:
				lstrcpy (plvdi->item.pszText,  "Tag");
				break;
			case Branch:
				lstrcpy (plvdi->item.pszText,  "Branch");
				break;
			}
			break;
        }
    }

	if (plvdi->item.mask & LVIF_IMAGE) {
		plvdi->item.iImage = type;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Helper functions

BOOL IsAnyItemSelected (HWND hwndListView)
{
    int nItems = ListView_GetItemCount (hwndListView);
    if (nItems > 0) {
        for (int i=0; i<nItems; i++) {
            if (ListView_GetItemState (hwndListView, i, LVIS_SELECTED) &
                LVIS_SELECTED)
                return TRUE;
        }
    }
    return FALSE;
}
