// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - May 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifdef _MSC_VER
#pragma warning(disable:4786 4503)
#endif

#include "PathUtils.h"
#include <windows.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "TortoiseDebug.h"
#include <shlobj.h>

bool FileExists(const char* filename)
{
	int attributes = GetFileAttributes(filename);
	return attributes == -1 ?  false: true;
}

bool IsDirectory(const char* filename)
{
	int attributes = GetFileAttributes(filename);
	if (attributes == -1)
		return false;

	return (attributes & FILE_ATTRIBUTE_DIRECTORY) != 0;
}

std::string GetDirectoryAbove(std::string filename)
{
	if (filename.empty())
		return filename;
	RemoveTrailingDelimiter(filename);

	// Return the directory above this file/directory
	std::string::size_type pos = filename.find_last_of("\\");
	if (pos == std::string::npos)
		return filename;
	return filename.substr(0, pos);

}

std::string GetDirectoryPart(std::string filename)
{
	if (filename.empty())
		return filename;
	RemoveTrailingDelimiter(filename);

	if (IsDirectory(filename.c_str()))
		return filename;

	// A file, so return the directory above it
	std::string::size_type pos = filename.find_last_of("\\");
	if (pos == std::string::npos)
		return filename;
	return filename.substr(0, pos);

}

void RemoveTrailingDelimiter(std::string& pathorfile)
{
	if (pathorfile.empty())
		return;

	// Remove trailing slash
	if (pathorfile[pathorfile.size() - 1] == '\\'
		|| pathorfile[pathorfile.size() - 1] == '/')
		pathorfile = pathorfile.substr(0, pathorfile.size() - 1);
}

void EnsureTrailingDelimiter(std::string& pathorfile)
{
	if (pathorfile.empty())
		return;

	if (pathorfile[pathorfile.size() - 1] == '\\' || pathorfile[pathorfile.size() - 1] == '/')
		return;

	pathorfile += "\\";
}

bool CVSFolderHere(std::string path)
{
	RemoveTrailingDelimiter(path);
	std::string CVSFolder = path + "\\CVS";
	// We also check for CVS/Repository, to minimise the chance of hitting
	// a folder called CVS that was made by the user or another program.
	std::string CVSFolderRepository = path + "\\CVS\\Repository";
	return FileExists(CVSFolder.c_str()) && FileExists(CVSFolderRepository.c_str());
}

std::string ExtractLastPart(std::string pathorfile)
{
	RemoveTrailingDelimiter(pathorfile);

	std::string::size_type pos = pathorfile.find_last_of("\\");
	if (pos == std::string::npos)
		return pathorfile;
	return pathorfile.substr(pos + 1, pathorfile.size());

}

int FileSize(const char* filename)
{
	struct stat buf;
	int i = stat( filename, &buf );
	if (i == -1)
		return -1;
	return buf.st_size;
}

std::string GetTemporaryPath()
{
	char tempDir[_MAX_PATH + 1];
	if (GetTempPath(_MAX_PATH, tempDir) == 0)
	{
		std::string message = "Failed to find temporary path";
		MessageBox(0, message.c_str(), "TortoiseCVS", MB_OK);
		// TODO: Better general error handling - throw exception here instead
	}
	return std::string(tempDir);
}

std::string UniqueTemporaryFile()
{
	char tempFile[_MAX_PATH + 1];
	if (!GetTempFileName(GetTemporaryPath().c_str(), "TCV", 0, tempFile))
	{
		std::string message = "Failed to generate temporary filename";
		MessageBox(0, message.c_str(), "TortoiseCVS", MB_OK);
		// TODO: Better general error handling - throw exception here instead
	}
	return std::string(tempFile);
}

static void ItemListFree(LPITEMIDLIST pidl)
{
    if ( pidl )
    {
        LPMALLOC pMalloc;
        SHGetMalloc(&pMalloc);
        if ( pMalloc )
        {
            pMalloc->Free(pidl);
            pMalloc->Release();
        }
        else
        {
            ASSERT(false);
        }
    }
}

// We don't use SHGetSpecialFolderPath as older versions
// of Windows NT don't support it.
std::string MyDocumentsFolder()
{
    HRESULT hr;
    ITEMIDLIST *pidl;
    hr = SHGetSpecialFolderLocation(NULL, CSIDL_PERSONAL, &pidl);
    if (FAILED(hr))
        return "";

    static char dir[MAX_PATH];
    if (!SHGetPathFromIDList(pidl, dir))
    {
	    ItemListFree(pidl);
		return "";
    }

    ItemListFree(pidl);
	return dir;
}


void GetRecursiveContents(std::string path, std::vector<std::string>& files, const std::string& wildcard)
{
	EnsureTrailingDelimiter(path);

	std::string str = path + wildcard;
	WIN32_FIND_DATA finddata;
	HANDLE h = FindFirstFile( str.c_str(), &finddata );
	if( h == INVALID_HANDLE_VALUE )
		// TODO: Error handling
		// Error::BrownTrousersTime("Couldn't scan directory %s", str.c_str());
		return ;
	do
	{
		if (std::string(finddata.cFileName) != "." && std::string(finddata.cFileName) != "..")
		{
			// Add files and directories
			files.push_back( path + finddata.cFileName  );

			// Recurse into directories
			if( (finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) )
				GetRecursiveContents(path + finddata.cFileName, files, wildcard);
		}
	}
	while( FindNextFile( h, &finddata ) );
	FindClose( h );

	// TODO: Error handling
	// if( err != ERROR_NO_MORE_FILES )
	//	Error::BrownTrousersTime("Error scanning directory %s", str.c_str());
}

// Return true if PathName is a drive letter ("A:")
// or a root path ("A:\"), false otherwise.
bool IsRootPath(const std::string& PathName)
{
	bool Result = false;
	
	if ((PathName.length() >= 2) && (PathName[1] == ':'))
	{
		if ((PathName.length() == 2) ||
			((PathName.length() == 3) && (PathName[2] == '\\')))
			Result = true;
	}

	// only eliminate removable drives
	// TODO[JR]: fix horrible removable drive hack
	if (Result)
	{
		switch (PathName[0])
		{
		case 'a': case 'A':
		case 'b': case 'B':
			break;
		default:
			Result = false;
			break;
		}
	}

	return Result;
}

std::string GetFirstPart(std::string pathorfile)
{
	std::string::size_type pos = pathorfile.find_first_of("\\");
	if (pos == std::string::npos)
		return "";
	return pathorfile.substr(0, pos + 1);
}

std::string StripFirstPart(std::string pathorfile)
{
	std::string::size_type pos = pathorfile.find_first_of("\\");
	if (pos == std::string::npos)
		return pathorfile;
	return pathorfile.substr(pos + 1, pathorfile.size());
}

bool CopyFolderRecursively(std::string source, std::string dest)
{
	// Double-null terminated strings required
	source += (char)0;
	source += (char)0;
	dest += (char)0;
	dest += (char)0;

	SHFILEOPSTRUCT fileOp;
	fileOp.hwnd = 0;
	fileOp.wFunc = FO_COPY;
	fileOp.pFrom = source.c_str();
	fileOp.pTo = dest.c_str();
	fileOp.fFlags = FOF_NOCONFIRMMKDIR;

	if( SHFileOperation( &fileOp ) != 0 )
		return false;

	return true;
}
