// TortoiseCVS - a Windows shell extension for easy version control

// Copyright (C) 2000 - Francis Irving
// <francis@flourish.org> - May 2000

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "TortoiseRegistry.h"
#include <windows.h>
#include <stdio.h>

// Note the lack of a trailing "\" - Windows 95 doesn't like one
const char tortoisePath[] = "Software\\TortoiseCVS";
const char wincvsPath[] = "Software\\WinCVS\\wincvs\\CVS settings";

std::string TortoiseRegistry::ReadString(const std::string& name, const std::string& defaultValue)
{
	HKEY key;
	if( RegOpenKeyEx( HKEY_CURRENT_USER, tortoisePath, 0, KEY_READ, &key )
		!= ERROR_SUCCESS )
	{
		return defaultValue;
	}

	char buf[2048];
	DWORD len = sizeof( buf );
	DWORD type;
	if( RegQueryValueEx( key, name.c_str(), NULL, &type, (unsigned char*)buf, &len )
		!= ERROR_SUCCESS || type != REG_SZ)
	{
		RegCloseKey(key);
		return defaultValue;
	}

	RegCloseKey(key);
	return std::string(buf);
}

void TortoiseRegistry::WriteString(const std::string& name, const std::string& value)
{
	HKEY key;

	if( RegCreateKeyEx( HKEY_CURRENT_USER, tortoisePath, 0,
		NULL, 0, KEY_WRITE, NULL, &key, NULL ) != ERROR_SUCCESS )
	{
		return;
	}

	RegSetValueEx( key, name.c_str(), 0,
		REG_SZ, (unsigned char*)value.c_str(), value.size() + 1 ); // include NULL in length for WinNT

	RegCloseKey(key);
}

int TortoiseRegistry::ReadInteger(const std::string& name, int defaultValue)
{
	HKEY key;
	if( RegOpenKeyEx( HKEY_CURRENT_USER, tortoisePath, 0, KEY_READ, &key )
		!= ERROR_SUCCESS )
	{
		return defaultValue;
	}

	DWORD value;
	DWORD len = sizeof( value );
	DWORD type;
	if( RegQueryValueEx( key, name.c_str(), NULL, &type, (unsigned char*)&value, &len )
		!= ERROR_SUCCESS || type != REG_DWORD)
	{
		RegCloseKey(key);
		return defaultValue;
	}

	RegCloseKey(key);
	return value;
}

void TortoiseRegistry::WriteInteger(const std::string& name, int value)
{
	HKEY key;

	if( RegCreateKeyEx( HKEY_CURRENT_USER, tortoisePath, 0,
		NULL, 0, KEY_WRITE, NULL, &key, NULL ) != ERROR_SUCCESS )
	{
		return;
	}

	RegSetValueEx( key, name.c_str(), 0,
		REG_DWORD, (unsigned char*)&value, sizeof(value) );

	RegCloseKey(key);
}

bool TortoiseRegistry::ReadBoolean(const std::string& name, bool defaultValue)
{
	int value = ReadInteger(name, defaultValue ? 1 : 0);
	return (value == 1) ? true : false;
}

void TortoiseRegistry::WriteBoolean(const std::string& name, bool value)
{
	WriteInteger(name, value ? 1 : 0);
}

// TODO: create a separate subkey for lists
void TortoiseRegistry::ReadList( std::string const& basekey,
	std::list< std::string >& strings )
{
	char keyname[ MAX_PATH ];
	int i=0;
	while( i<99 )
	{
		sprintf( keyname, "%s %d", basekey.c_str(), i );
		std::string str = TortoiseRegistry::ReadString( keyname );
		if( str.empty() )
			break;
		strings.push_back( str );
		++i;
	}
}


void TortoiseRegistry::WriteList( std::string const& basekey,
	std::list< std::string > const& strings )
{
	char keyname[ MAX_PATH ];
	std::list< std::string >::const_iterator it;;
	int i=0;
	for( it=strings.begin(); it != strings.end(); ++it )
	{
		sprintf( keyname, "%s %d", basekey.c_str(), i++ );
		TortoiseRegistry::WriteString( keyname, *it );
	}
	// TODO: clear old list entries
}


std::string TortoiseRegistry::ReadWinCVSString(const std::string& name, const std::string& defaultValue)
{
    HKEY key;
    LONG retval = RegOpenKeyEx(HKEY_CURRENT_USER, wincvsPath, 0, KEY_QUERY_VALUE, &key);
    if (retval != ERROR_SUCCESS)
		return defaultValue;

    TCHAR data[MAX_PATH];
    DWORD datasize = MAX_PATH;
	DWORD type;
	if( RegQueryValueEx(key, name.c_str(), NULL, &type, (unsigned char*)data, &datasize)
		!= ERROR_SUCCESS || type != REG_BINARY)
	{
		RegCloseKey(key);
		return defaultValue;
	}
	RegCloseKey(key);

    return std::string(data);
}


