%{
/*
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 1, or (at your option)
** any later version.

** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.

** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "common.h"
#include "CvsLogParser.h"

#if qCvsDebug
#	define YYDEBUG 1
#else
#	define YYDEBUG 0
#endif

static void expect(kTokens type);

static int sInt;
static CLogStr sStr;
static CRevNumber sRev;

// owns all the parse actions
CvsLogParser * gLog;

%}

%union {
	char *string;
	int aint;
};

%token <aint> INTEGER
%token <string> STRING
%token <string> TAG
%token <revnum> REVNUM
%token <aint> ID_CHAR
%token <aint> error

%token ID_DESCRIPTION
%token ID_TOTREVISIONS
%token ID_SELREVISIONS
%token ID_KEYWORD
%token ID_SYMBOLIC
%token ID_ACCESS
%token ID_LOCKS
%token ID_STRICT
%token ID_BRANCH
%token ID_HEAD
%token ID_WORKING
%token ID_RCS
%token ID_REVISION
%token ID_DATE
%token ID_BRANCHES
%token ID_REVFOLLOW
%token ID_ENDFOLLOW
%token ID_LINES
%token ID_ANYBUTSEP
%token ID_CHAR
%token ID_LOCKEDBY
%token ID_AUTHOR
%token ID_STATE

%%

//*****************************************************************************
// multi-file CVS log output

cvslog:		allDeclarations
	;

allDeclarations:
		/* Empty is OK */
	|	allDeclarations allDeclaration
	;

	allDeclaration:
		rcsDeclarations
	|	ID_REVFOLLOW				{  gLog->RevFile().StartParse(); }
			allrevDeclarations			{ gLog->RevFile().EndParse(); }
	|	ID_ENDFOLLOW				{ gLog->RcsFile().EndParse(); expect(kNone); }
	|	error 						{
										int c = $1;
										yyclearin;
										expect(kAnyButSep);

										if(gLog->RevFile().IsParsing())
											gLog->RevFile()->DescLog() << (char)c;
										else if(gLog->RcsFile().IsParsing())
											gLog->RcsFile()->DescLog() << (char)c;
									}
			anyDeclarations
	|	tagDeclaration				{
										expect(kAnyButSep);
										if(gLog->RevFile().IsParsing())
											gLog->RevFile()->DescLog() << sStr;
										else if(gLog->RcsFile().IsParsing())
											gLog->RcsFile()->DescLog() << sStr;
									}
			anyDeclarations
	;

//*****************************************************************************
// RCS header declarations for one file

rcsDeclarations:
	|	rcsDeclarations rcsDeclaration
	;

rcsDeclaration:
		rcsFileDeclaration
	|	workingFileDeclaration
	|	headRevDeclaration
	|	branchRevDeclaration
	|	locksDeclaration
	|	accessDeclaration
	|	symbolicNamesDeclaration
	|	keywordDeclaration
	|	totRevDeclaration
	|	descDeclaration
	;

anyDeclarations:
		/* Empty is OK */
	|	anyDeclarations anyDeclaration
	;

anyDeclaration:
		/* Empty is OK */
	|	ID_CHAR			{
							char c = $1;
							if(gLog->RevFile().IsParsing())
								gLog->RevFile()->DescLog() << c;
							else if(gLog->RcsFile().IsParsing())
								gLog->RcsFile()->DescLog() << c;
						}
	|	STRING			{
							if(gLog->RevFile().IsParsing())
								gLog->RevFile()->DescLog() << $1;
							else if(gLog->RcsFile().IsParsing())
								gLog->RcsFile()->DescLog() << $1;
							free($1);
						}
	;

descDeclaration:
		ID_DESCRIPTION ':'					{ expect(kAnyButSep); }
			descListDeclarations
	;

descListDeclarations:
		/* Empty is OK */
	|	descListDeclarations descListDeclaration
	;

descListDeclaration:
		/* Empty is OK */
	|	ID_CHAR			{
							char c = $1;
							if(gLog->RcsFile().IsParsing())
								gLog->RcsFile()->DescLog() << c;
						}
	|	STRING			{
							if(gLog->RcsFile().IsParsing())
								gLog->RcsFile()->DescLog() << $1;
							free($1);
						}
	;

//*****************************************************************************
// one revision of a file

allrevDeclarations:
	|	allrevDeclarations allrevDeclaration
	;

allrevDeclaration:
		ID_REVISION 				{ expect(kRevnum); }
			REVNUM					{ gLog->RevFile()->RevNum() = sRev; sRev.reset(); }
	|	ID_DATE	':'						{ expect(kDate); }
			intDeclaration '/'		{ gLog->RevFile()->RevTime().tm_year = sInt - 1900; }
			intDeclaration '/'		{ gLog->RevFile()->RevTime().tm_mon = sInt; }
			intDeclaration			{ gLog->RevFile()->RevTime().tm_mday = sInt; }
			intDeclaration ':'		{ gLog->RevFile()->RevTime().tm_hour = sInt; }
			intDeclaration ':'		{ gLog->RevFile()->RevTime().tm_min = sInt; }
			intDeclaration			{ gLog->RevFile()->RevTime().tm_sec = sInt; expect(kNone); }
			';'
	|	ID_BRANCHES ':'		{ expect(kRevnum); }
			revnumListDeclarations
	|	ID_AUTHOR ':'
			tagDeclaration ';'		{ gLog->RevFile()->Author() = sStr; }
	|	ID_LOCKEDBY ':'
			tagDeclaration ';'		{ gLog->RevFile()->Locker() = sStr; }
	|	ID_STATE ':'
			tagDeclaration ';'		{ gLog->RevFile()->State() = sStr; }
	|	ID_LINES ':'					{ expect(kInt); }
			intDeclaration			{ expect(kInt); gLog->RevFile()->ChgPos() = sInt; }
			intDeclaration			{ gLog->RevFile()->ChgNeg() = sInt; }
	;

revnumListDeclarations:
		/* Empty is OK */
	|	revnumListDeclarations revnumListDeclaration
	;

revnumListDeclaration:
		/* Empty is OK */
	|	REVNUM ';'				{
				gLog->Branch().StartParse();
				*gLog->Branch() = sRev;
				gLog->Branch().EndParse();
				sRev.reset();
				expect(kRevnum);
			}
	;

//*****************************************************************************
// basic syntax units

stringDeclaration:
		STRING				{ sStr = $1; free($1); }
	;

tagDeclaration:
		TAG					{ sStr = $1; free($1); }
	;

intDeclaration:
		INTEGER			{ sInt = $1; }
	;

//*****************************************************************************
// RCS header declarations for one file

totRevDeclaration:
		ID_TOTREVISIONS ':'			{ expect(kInt); }
			intDeclaration			{ gLog->RcsFile()->TotRevisions() = sInt; }
			selRevDeclaration
	;

selRevDeclaration:
		/* Empty is OK */
	|	';' ID_SELREVISIONS ':'		{ expect(kInt); }
			intDeclaration			{ gLog->RcsFile()->SelRevisions() = sInt; }

keywordDeclaration:
		ID_KEYWORD ':'			{ expect(kString); }
			stringDeclaration	{ gLog->RcsFile()->KeywordSubst() = sStr; }
	;

symbolicNamesDeclaration:
		ID_SYMBOLIC ':' symbolicListDeclarations
	;

symbolicListDeclarations:
		/* Empty is OK */
	|	symbolicListDeclarations symbolicListDeclaration
	;

symbolicListDeclaration:
		tagDeclaration ':'		{
									expect(kRevnum);
									gLog->Symbolic().StartParse();
									gLog->Symbolic()->Tag() = sStr;
								}
			REVNUM				{
									sRev.Tag() = gLog->Symbolic()->Tag();
									*gLog->Symbolic() = sRev;
									gLog->Symbolic().EndParse();
									sRev.reset();
								}
	;

accessDeclaration:
		ID_ACCESS ':' accessListDeclarations
	;

accessListDeclarations:
		/* Empty is OK */
	|	accessListDeclarations accessListDeclaration
	;

accessListDeclaration:
		tagDeclaration ':'		{
									gLog->Access().StartParse();
									*gLog->Access() = sStr;
									gLog->Access().EndParse();
								}
		tagDeclaration
	|	tagDeclaration			{
									gLog->Access().StartParse();
									*gLog->Access() = sStr;
									gLog->Access().EndParse();
								}
	;

locksDeclaration:
		ID_LOCKS ':'					{ gLog->RcsFile()->LockStrict() = false; }
			locksListDeclarations
	|	ID_LOCKS ':' ID_STRICT			{ gLog->RcsFile()->LockStrict() = true; }
			locksListDeclarations
	;

locksListDeclarations:
		/* Empty is OK */
	|	locksListDeclarations locksListDeclaration
	;

locksListDeclaration:
		tagDeclaration ':'		{
									expect(kRevnum);
									gLog->Lock().StartParse();
									gLog->Lock()->Tag() = sStr;
								}
			REVNUM				{
									sRev.Tag() = gLog->Lock()->Tag();
									*gLog->Lock() = sRev;
									gLog->Lock().EndParse();
									sRev.reset();
								}
	;

branchRevDeclaration:
		ID_BRANCH ':'		{ expect(kRevnumOrNil); }
			REVNUM				{ gLog->RcsFile()->BranchRev() = sRev; sRev.reset(); }
	;

headRevDeclaration:
		ID_HEAD ':'			{ expect(kRevnum); }
			REVNUM 			{ gLog->RcsFile()->HeadRev() = sRev; sRev.reset(); }
	;

workingFileDeclaration:
		ID_WORKING ':'			{ expect(kString); }
			stringDeclaration	{ gLog->RcsFile()->WorkingFile() = sStr; }
	;

rcsFileDeclaration:
		ID_RCS ':'						{ expect(kString); gLog->RcsFile().StartParse(); }
			stringDeclaration	{ gLog->RcsFile()->RcsFile() = sStr; }
	;

%%

void yyerror(const char * /*txt*/, ...)
{
	expect(kAnyButSep);
}

static void expect(kTokens type)
{
	expectToken = type;
}

void yylval_set(const char *txt)
{
	yylval.string = strdup(txt);
}

void yylval_set(int anint)
{
	yylval.aint = anint;
}

void yylval_set(const CRevNumber & rev)
{
	sRev = rev;
}

void yyreset(void)
{
	currentLineNumber = 1;
	sStr = "";
	sInt = 0;
	sRev.reset();
	gLog->Reset();
	expect(kNone);
}
